import numpy as np

import astropy.units as u
import astropy.wcs
from astropy.coordinates import SkyCoord

from sunpy import log
from sunpy.coordinates import frames, sun
from sunpy.util import MetaDict

__all__ = ['make_fitswcs_header', 'get_observer_meta', 'make_heliographic_header', 'make_hpr_header']


@u.quantity_input(equivalencies=u.spectral())
def make_fitswcs_header(data,
                        coordinate,
                        reference_pixel: u.pix = None,
                        scale: u.arcsec/u.pix = None,
                        rotation_angle: u.deg = None,
                        rotation_matrix=None,
                        instrument=None,
                        telescope=None,
                        observatory=None,
                        detector=None,
                        wavelength: u.angstrom = None,
                        exposure: u.s = None,
                        projection_code="TAN",
                        unit=None):
    """
    Function to create a FITS-WCS header from a coordinate object
    (`~astropy.coordinates.SkyCoord`) that is required to
    create a `~sunpy.map.GenericMap`.

    Parameters
    ----------
    data : `~numpy.ndarray`, `~astropy.units.Quantity`, or `tuple`
        Array data of Map for which a header is required, or the shape of the
        data array (in numpy order, i.e. ``(y_size, x_size)``).
    coordinate : `~astropy.coordinates.SkyCoord` or `~astropy.coordinates.BaseCoordinateFrame`
        The coordinate of the reference pixel.
    reference_pixel : `~astropy.units.Quantity`, optional
        Reference pixel along each axis. These are expected to be Cartestian ordered, i.e
        the first index is the x axis, second index is the y axis. Defaults to
        the center of data array, ``(data.shape[1] - 1)/2., (data.shape[0] - 1)/2.)``,
        this argument is zero indexed (Python convention) not 1 indexed (FITS
        convention).
    scale : `~astropy.units.Quantity` of size 2, optional
        Pixel scaling along x and y axis (i.e. the spatial scale of the pixels (dx, dy)). These are
        expected to be Cartestian ordered, i.e [dx, dy].
        Defaults to ``([1., 1.] arcsec/pixel)``.
    rotation_angle : `~astropy.units.Quantity`, optional
        Coordinate system rotation angle, will be converted to a rotation
        matrix and stored in the ``PCi_j`` matrix. Can not be specified with
        ``rotation_matrix``. Defaults to no rotation.
    rotation_matrix : `~numpy.ndarray` of dimensions 2x2, optional
        Matrix describing the rotation required to align solar North with
        the top of the image in FITS ``PCi_j`` convention. Can not be specified
        with ``rotation_angle``.
    instrument : `~str`, optional
        Name of the instrument of the observation.
    telescope : `~str`, optional
        Name of the telescope of the observation.
    observatory : `~str`, optional
        Name of the observatory of the observation.
    detector : `str`, optional
        Name of the detector of the observation.
    wavelength : `~astropy.units.Quantity`, optional
        Wavelength of the observation as an astropy quantity, e.g. 171*u.angstrom.
        From this keyword, the meta keywords ``wavelnth`` and ``waveunit`` will be populated.
    exposure : `~astropy.units.Quantity`, optional
        Exposure time of the observation
    projection_code : `str`, optional
        The FITS standard projection code for the new header.
    unit : `~astropy.units.Unit`, optional
        Units of the array data of the Map. This will populate the the ``'bunit'`` meta keyword.
        If ``data`` is a `~astropy.units.Quantity`, the unit specified here will take precedence
        over the unit information attached to ``data``.

    Returns
    -------
    `~sunpy.util.MetaDict`
        The header information required for making a `sunpy.map.GenericMap`.

    Notes
    -----
    The observer coordinate is taken from the observer property of the ``reference_pixel``
    argument.

    Examples
    --------
    >>> import sunpy.map
    >>> from sunpy.coordinates import frames
    >>> from astropy.coordinates import SkyCoord
    >>> import astropy.units as u
    >>> import numpy as np

    >>> data = np.random.rand(1024, 1024)
    >>> my_coord = SkyCoord(0*u.arcsec, 0*u.arcsec, obstime="2017-08-01",
    ...                     observer = 'earth', frame=frames.Helioprojective)
    >>> my_header = sunpy.map.make_fitswcs_header(data, my_coord)
    >>> my_map = sunpy.map.Map(data, my_header)
    """
    coordinate = _validate_coordinate(coordinate)

    if hasattr(data, "shape"):
        shape = data.shape
    else:
        shape = data

    if hasattr(data, "unit"):
        if unit is None:
            unit = data.unit
        else:
            log.info("Overwriting data's current unit with specified unit.")

    meta_wcs = _get_wcs_meta(coordinate, projection_code)

    meta_wcs = _set_instrument_meta(meta_wcs, instrument, telescope, observatory, detector, wavelength, exposure, unit)
    meta_wcs = _set_transform_params(meta_wcs, coordinate, reference_pixel, scale, shape)
    meta_wcs = _set_rotation_params(meta_wcs, rotation_angle, rotation_matrix)

    if getattr(coordinate, 'observer', None) is not None:
        # Have to check for str, as doing == on a SkyCoord and str raises an error
        if isinstance(coordinate.observer, str) and coordinate.observer == 'self':
            dsun_obs = coordinate.radius
        else:
            dsun_obs = coordinate.observer.radius
        meta_wcs['rsun_obs'] = sun._angular_radius(coordinate.rsun, dsun_obs).to_value(u.arcsec)

    meta_dict = MetaDict(meta_wcs)
    return meta_dict


def _validate_coordinate(coordinate):
    if not isinstance(coordinate, SkyCoord | frames.BaseCoordinateFrame):
        raise ValueError("coordinate needs to be a coordinate frame or an SkyCoord instance.")

    if isinstance(coordinate, SkyCoord):
        coordinate = coordinate.frame

    if coordinate.obstime is None:
        raise ValueError("The coordinate needs an observation time, `obstime`.")

    if isinstance(coordinate, frames.Heliocentric):
        raise ValueError("This function does not currently support heliocentric coordinates.")
    return coordinate


def _set_rotation_params(meta_wcs, rotation_angle, rotation_matrix):
    if rotation_angle is not None and rotation_matrix is not None:
        raise ValueError("Can not specify both rotation angle and rotation matrix.")
    if rotation_angle is None and rotation_matrix is None:
        rotation_angle = 0 * u.deg

    if rotation_angle is not None:
        lam = meta_wcs['cdelt2'] / meta_wcs['cdelt1']
        p = np.deg2rad(rotation_angle)

        rotation_matrix = np.array([[np.cos(p), -1 * lam * np.sin(p)],
                                    [1/lam * np.sin(p), np.cos(p)]])

    if rotation_matrix is not None:
        (meta_wcs['PC1_1'], meta_wcs['PC1_2'],
         meta_wcs['PC2_1'], meta_wcs['PC2_2']) = (rotation_matrix[0, 0], rotation_matrix[0, 1],
                                                  rotation_matrix[1, 0], rotation_matrix[1, 1])
    return meta_wcs


def _set_transform_params(meta_wcs, coordinate, reference_pixel, scale, shape):
    meta_wcs['naxis'] = 2
    meta_wcs['naxis1'] = shape[1]
    meta_wcs['naxis2'] = shape[0]

    if reference_pixel is None:
        reference_pixel = u.Quantity([(shape[1] - 1)/2.*u.pixel,
                                      (shape[0] - 1)/2.*u.pixel])
    if scale is None:
        scale = [1., 1.] * (u.arcsec/u.pixel)

    meta_wcs['crval1'], meta_wcs['crval2'] = (coordinate.spherical.lon.to_value(meta_wcs['cunit1']),
                                              coordinate.spherical.lat.to_value(meta_wcs['cunit2']))

    # When the native latitude of the fiducial point is 0 degrees, which is typical for cylindrical
    # projections, the correct value of `lonpole` depends on whether the world latitude of the
    # fiducial point is greater than or less than its native latitude.
    if coordinate.spherical.lat.to_value(u.deg) < meta_wcs['theta0']:
        meta_wcs['LONPOLE'] = 180.
    else:
        meta_wcs['LONPOLE'] = 0.
    del meta_wcs['theta0']  # remove the native latitude of the fiducial point

    # Add 1 to go from input 0-based indexing to FITS 1-based indexing
    meta_wcs['crpix1'], meta_wcs['crpix2'] = (reference_pixel[0].to_value(u.pixel) + 1,
                                              reference_pixel[1].to_value(u.pixel) + 1)

    meta_wcs['cdelt1'] = scale[0].to_value(u.Unit(meta_wcs['cunit1']) / u.pixel)
    meta_wcs['cdelt2'] = scale[1].to_value(u.Unit(meta_wcs['cunit2']) / u.pixel)
    return meta_wcs


def _get_wcs_meta(coordinate, projection_code):
    """
    Function to get WCS meta from the SkyCoord using
    `astropy.wcs.utils.celestial_frame_to_wcs`

    Parameters
    ----------
    coordinate : `~astropy.coordinates.BaseCoordinateFrame`

    Returns
    -------
    `dict`
        Containing the WCS meta information
            * ctype1, ctype2
            * cunit1, cunit2
            * date_obs
            * observer auxiliary information, if set on `coordinate`
    """

    coord_meta = {}

    skycoord_wcs = astropy.wcs.utils.celestial_frame_to_wcs(coordinate, projection_code)

    cunit1, cunit2 = skycoord_wcs.wcs.cunit
    coord_meta = dict(skycoord_wcs.to_header())
    coord_meta['cunit1'], coord_meta['cunit2'] = cunit1.to_string("fits"), cunit2.to_string("fits")
    coord_meta['theta0'] = skycoord_wcs.wcs.theta0  # add the native latitude of the fiducial point

    return coord_meta


@u.quantity_input
def get_observer_meta(observer, rsun: (u.Mm, None) = None):
    """
    Function to get observer meta from coordinate frame.

    Parameters
    ----------
    observer : `~astropy.coordinates.BaseCoordinateFrame`
        The coordinate of the observer, must be transformable to Heliographic
        Stonyhurst.
    rsun : `astropy.units.Quantity`, optional
        The radius of the Sun. If ``None``, the RSUN_OBS and RSUN_REF keys are
        not set.

    Returns
    -------
    coord_meta : `dict`
        WCS metadata, with the keys ``['hgln_obs', 'hglt_obs', 'dsun_obs']``,
        and additionally if ``rsun`` is given ``['rsun_obs', 'rsun_ref']``.
    """
    observer = observer.transform_to(frames.HeliographicStonyhurst(obstime=observer.obstime))
    coord_meta = {}

    coord_meta['hgln_obs'] = observer.lon.to_value(u.deg)
    coord_meta['hglt_obs'] = observer.lat.to_value(u.deg)
    coord_meta['dsun_obs'] = observer.radius.to_value(u.m)
    if rsun is not None:
        coord_meta['rsun_ref'] = rsun.to_value(u.m)
        coord_meta['rsun_obs'] = sun._angular_radius(rsun, observer.radius).to_value(u.arcsec)

    return coord_meta


def _set_instrument_meta(meta_wcs, instrument, telescope, observatory, detector, wavelength, exposure, unit):
    """
    Function to correctly name keywords from keyword arguments
    """
    if instrument is not None:
        meta_wcs['instrume'] = str(instrument)
    if telescope is not None:
        meta_wcs['telescop'] = str(telescope)
    if observatory is not None:
        meta_wcs['obsrvtry'] = str(observatory)
    if detector is not None:
        meta_wcs['detector'] = str(detector)
    if wavelength is not None:
        meta_wcs['wavelnth'] = wavelength.to_value()
        meta_wcs['waveunit'] = wavelength.unit.to_string("fits")
    if exposure is not None:
        meta_wcs['exptime'] = exposure.to_value(u.s)
    if unit is not None:
        _unit = u.Unit(unit)
        # NOTE: This conditional is because 'DN' is not part of the
        # FITS standard and thus cannot be serialized to a string
        # using the FITS format
        unit_format = 'generic' if u.DN in _unit.bases else 'fits'
        meta_wcs['bunit'] = _unit.to_string(format=unit_format)
    return meta_wcs


@u.quantity_input
def make_heliographic_header(date, observer_coordinate, shape, *, frame, projection_code="CAR",
                             map_center_longitude: u.Quantity[u.deg] = 0.0*u.deg):
    """
    Construct a FITS-WCS header for a full-Sun heliographic (Carrington or Stonyhurst) coordinate frame.

    The date-time and observer coordinate of the new coordinate frame
    are taken from the input map. The resulting WCS covers the full surface
    of the Sun, and has a reference coordinate at (0, 0) degrees Longitude/Latitude.

    Parameters
    ----------
    date :
        Date for the output header.
    observer_coordinate :
        Observer coordinate for the output header.
    shape : [int, int]
        Output map shape, number of pixels in (latitude, longitude).
    frame : {'carrington', 'stonyhurst'}
        Coordinate frame.
    projection_code : {'CAR', 'CEA'}
        Projection to use for the latitude coordinate.
    map_center_longitude : `~astropy.units.Quantity`
        Heliographic longitude of the map center

    Returns
    -------
    `~sunpy.util.MetaDict`

    See Also
    --------
    sunpy.map.header_helper.make_fitswcs_header : A more generic header helper that can be used if more customisation is required.

    Examples
    --------
    >>> from sunpy.map.header_helper import make_heliographic_header
    >>> from sunpy.coordinates import get_earth
    >>>
    >>> date = '2020-01-01 12:00:00'
    >>> observer = get_earth(date)
    >>> header = make_heliographic_header(date, observer, [90, 180], frame='carrington')
    >>> header
    MetaDict([('wcsaxes': '2')
    ('crpix1': '90.5')
    ('crpix2': '45.5')
    ('cdelt1': '2.0')
    ('cdelt2': '2.0')
    ('cunit1': 'deg')
    ('cunit2': 'deg')
    ('ctype1': 'CRLN-CAR')
    ('ctype2': 'CRLT-CAR')
    ('crval1': '0.0')
    ('crval2': '0.0')
    ('lonpole': '0.0')
    ('latpole': '90.0')
    ('mjdref': '0.0')
    ('date-obs': '2020-01-01T12:00:00.000')
    ('rsun_ref': '695700000.0')
    ('dsun_obs': '147096975776.97')
    ('hgln_obs': '0.0')
    ('hglt_obs': '-3.0011725838606')
    ('naxis': '2')
    ('naxis1': '180')
    ('naxis2': '90')
    ('pc1_1': '1.0')
    ('pc1_2': '-0.0')
    ('pc2_1': '0.0')
    ('pc2_2': '1.0')
    ('rsun_obs': '975.53984320334...

    .. minigallery:: sunpy.map.make_heliographic_header
    """
    valid_codes = {"CAR", "CEA"}
    if projection_code not in valid_codes:
        raise ValueError(f"projection_code must be one of {valid_codes}")

    valid_frames = {'carrington', 'stonyhurst'}
    if frame not in valid_frames:
        raise ValueError(f"frame must be one of {valid_frames}")

    frame_out = SkyCoord(
        map_center_longitude,
        0 * u.deg,
        frame=f"heliographic_{frame}",
        obstime=date,
        observer=observer_coordinate,
        rsun=getattr(observer_coordinate, "rsun", None),
    )

    if projection_code == "CAR":
        scale = [360 / int(shape[1]), 180 / int(shape[0])] * u.deg / u.pix
    elif projection_code == "CEA":
        # Using the cylindrical equal-area (CEA) projection,
        # scale needs to be to 180/pi times the sin(latitude) spacing
        # See Section 5.5, Thompson 2006
        scale = [
            360 / int(shape[1]),
            (180 / np.pi) / (int(shape[0]) / 2)
        ] * u.deg / u.pix

    header = make_fitswcs_header(shape, frame_out, scale=scale, projection_code=projection_code)
    return header


@u.quantity_input
def make_hpr_header(observer_coordinate, shape, theta_binsize: u.Quantity[u.arcsec], *,
                    theta_min: u.Quantity[u.arcsec] = 0.0*u.arcsec,
                    psi_center: u.Quantity[u.deg] = 180.0*u.deg):
    """
    Construct a FITS-WCS header for a helioprojective radial coordinate frame.

    The header uses the plate carrée projection ("CAR" in FITS-WCS) with the
    position angle (``psi``) on the horizontal axis and impact angle (``theta``) on
    the vertical axis. To accommodate the FITS-WCS machinery, the vertical axis is
    actually the declination (``delta``), which is the impact angle minus 90 degrees.

    Parameters
    ----------
    observer_coordinate :
        Observer coordinate, with corresponding observation time.
    shape : [int, int]
        Output map shape, number of pixels in (theta, psi).
    theta_binsize : `~astropy.units.Quantity`
        The size of each pixel in the impact-angle direction.
    theta_min : `~astropy.units.Quantity`
        The minimum impact angle. Defaults to 0 arcsec.
    psi_center : `~astropy.units.Quantity`
        The center of the map in position angle. Defaults to 180 degrees.

    Returns
    -------
    `~sunpy.util.MetaDict`

    See Also
    --------
    sunpy.map.header_helper.make_fitswcs_header : A more generic header helper that can be used if more customisation is required.
    sunpy.coordinates.HelioprojectiveRadial : The coordinate-frame class

    Examples
    --------
    >>> import astropy.units as u
    >>> from sunpy.coordinates import get_earth

    >>> observer = get_earth('2024-09-16 01:02:03')
    >>> make_hpr_header(observer, (180, 360), theta_binsize=10*u.arcsec)
    MetaDict([('wcsaxes': '2')
    ('crpix1': '180.5')
    ('crpix2': '32400.5')
    ('cdelt1': '1.0')
    ('cdelt2': '10.0')
    ('cunit1': 'deg')
    ('cunit2': 'arcsec')
    ('ctype1': 'HRLN-CAR')
    ('ctype2': 'HRLT-CAR')
    ('crval1': '180.0')
    ('crval2': '0.0')
    ('lonpole': '0.0')
    ('latpole': '90.0')
    ('mjdref': '0.0')
    ('date-obs': '2024-09-16T01:02:03.000')
    ('rsun_ref': '695700000.0')
    ('dsun_obs': '150399816551.15')
    ('hgln_obs': '0.0')
    ('hglt_obs': '7.1884873184853')
    ('naxis': '2')
    ('naxis1': '360')
    ('naxis2': '180')
    ('pc1_1': '1.0')
    ('pc1_2': '-0.0')
    ('pc2_1': '0.0')
    ('pc2_2': '1.0')
    ('rsun_obs': '954.1164393469891')])

    .. minigallery:: sunpy.map.make_hpr_header
    """
    # The reference coord must be on the equator (delta = 0 deg)
    reference_coord = SkyCoord(
        psi_center,
        0*u.deg,
        frame="helioprojectiveradial",
        obstime=observer_coordinate.obstime,
        observer=observer_coordinate,
        rsun=getattr(observer_coordinate, "rsun", None),
    )

    # Calculate the correct location for this reference coord, which will likely be far off the data array
    reference_pixel = [
        shape[1] / 2. - 0.5,
        ((90*u.deg - theta_min) / theta_binsize).to_value(u.one) - 0.5
    ] * u.pix

    scale = u.Quantity([360*u.deg / int(shape[1]), theta_binsize]) / u.pix

    header = make_fitswcs_header(shape, reference_coord, reference_pixel=reference_pixel,
                                 scale=scale, projection_code="CAR")
    return header



import matplotlib
import pytest

import astropy.units as u

import sunpy.visualization.colormaps as cm
import sunpy.visualization.colormaps.color_tables as ct
from sunpy.tests.helpers import figure_test


# Checks that colormaps are imported by MPL
def test_get_cmap():
    for cmap in cm.cmlist.keys():
        assert cm.cmlist[cmap] == matplotlib.colormaps[cmap]


def test_invalid_show_cmaps():
    with pytest.raises(
            KeyError, match='No color maps found for search term'):
        cm.show_colormaps(search='asdfghjkl')


@pytest.mark.parametrize(
    ('f', 'match'),
    [(ct.aia_color_table, 'Invalid AIA wavelength.'),
     (ct.eit_color_table, 'Invalid EIT wavelength.'),
     (ct.sswidl_lasco_color_table, 'Invalid LASCO number.'),
     (ct.sxt_color_table, 'Invalid SXT filter type number.'),
     (ct.cor_color_table, 'Invalid COR number.'),
     (ct.trace_color_table, 'Invalid TRACE filter waveband passed.'),
     (ct.sot_color_table, r'Invalid \(or not supported\) SOT type.'),
     (ct.iris_sji_color_table, 'Invalid IRIS SJI waveband.'),
     (ct.stereo_hi_color_table, 'Valid HI cameras are 1 and 2'),
     (ct.suvi_color_table, 'Invalid SUVI wavelength.')]
)
def test_invalid_wavelengths(f, match):
    # Check that accessing non-existent color table values raises a ValueError
    with pytest.raises(ValueError, match=match):
        f(-100*u.m)


@figure_test
def test_cmap_visual():
    cm.show_colormaps()



import pytest
from hypothesis import given, settings

import astropy.units as u

import sunpy.net.dataretriever.sources.goes as goes
from sunpy.net import attrs as a
from sunpy.net.dataretriever.client import QueryResponse
from sunpy.net.tests.strategies import time_attr
from sunpy.time import parse_time


@pytest.fixture
def suvi_client():
    return goes.SUVIClient()


@settings(max_examples=5)
@given(time_attr())
def test_can_handle_query(time):
    # Don't use the fixture, as hypothesis complains
    suvi_client = goes.SUVIClient()
    assert suvi_client._can_handle_query(time, a.Instrument.suvi) is True
    assert suvi_client._can_handle_query(time, a.Instrument.suvi, a.Wavelength(131 * u.Angstrom)) is True
    assert suvi_client._can_handle_query(time) is False
    assert suvi_client._can_handle_query(time, a.Instrument.aia) is False
    assert suvi_client._can_handle_query(time, a.Instrument.suvi, a.goes.SatelliteNumber(16))


def mock_query_object(suvi_client):
    obj = {
        'Start Time': parse_time('2019/05/25 00:50'),
        'End Time': parse_time('2019/05/25 00:52'),
        'Instrument': 'SUVI',
        'Physobs': 'flux',
        'Source': 'GOES',
        'Provider': 'NOAA',
        'Level': '2',
        'Wavelength': 94 * u.Angstrom,
        'url': 'https://mock.url/suvi-l2-ci094.fits'
    }
    return QueryResponse([obj], client=suvi_client)


@pytest.mark.remote_data
@pytest.mark.parametrize(('start', 'end', 'wave', 'level', 'expected_num_files'), [
    ('2019/05/25 00:50', '2019/05/25 00:54', 94, '1b', 6),
    ('2019/05/25 00:50', '2019/05/25 00:54', 304, '2', 1),
])
def test_combined_search(suvi_client, start, end, wave, level, expected_num_files):
    goes_sat = a.goes.SatelliteNumber.sixteen
    qresponse = suvi_client.search(a.Time(start, end), a.Wavelength(wave * u.Angstrom), goes_sat, a.Level(level))
    assert len(qresponse) == expected_num_files


@pytest.mark.remote_data
def test_get_all_wavelengths_level2(suvi_client):
    """Check retrieval for all wavelengths without specifying one."""
    qresponse = suvi_client.search(a.Time('2019/05/25 00:50', '2019/05/25 00:52'),
                                   a.goes.SatelliteNumber.sixteen, a.Level(2))
    assert len(qresponse) == 6
    wavelengths = [w.value for w in qresponse['Wavelength']]
    expected_wavelengths = {94, 131, 171, 195, 284, 304}
    assert set(wavelengths) == expected_wavelengths


@pytest.mark.remote_data
def test_fetch_real_file(suvi_client, tmp_path):
    start = '2019/05/25 00:50'
    end = '2019/05/25 00:52'
    wave = 94 * u.Angstrom
    goes_sat = a.goes.SatelliteNumber.sixteen
    tr = a.Time(start, end)
    download_list = suvi_client.fetch(
        suvi_client.search(tr, a.Instrument.suvi, a.Wavelength(wave), goes_sat, a.Level(2)),
        path=str(tmp_path)
    )
    assert len(download_list) == len(
        suvi_client.search(tr, a.Instrument.suvi, a.Wavelength(wave), goes_sat, a.Level(2))
    )


@pytest.mark.remote_data
def test_fetch_mocked_file(suvi_client, mocker):
    start = '2019/05/25 00:50'
    end = '2019/05/25 00:52'
    wave = 94 * u.Angstrom
    goes_sat = a.goes.SatelliteNumber.sixteen
    tr = a.Time(start, end)
    mocker.patch.object(suvi_client, 'search', return_value=mock_query_object(suvi_client))
    mocker.patch.object(suvi_client, 'fetch', return_value=['mockfile.fits'])
    download_list = suvi_client.fetch(
        suvi_client.search(tr, a.Instrument.suvi, a.Wavelength(wave), goes_sat, a.Level(2))
    )
    assert download_list == ['mockfile.fits']
    assert len(download_list) == len(
        suvi_client.search(tr, a.Instrument.suvi, a.Wavelength(wave), goes_sat, a.Level(2))
    )

def test_attr_reg():
    a.Instrument.suvi = a.Instrument("SUVI")
    a.goes.SatelliteNumber.A16 = a.goes.SatelliteNumber("16")


def test_show(suvi_client):
    mock_qr = mock_query_object(suvi_client)
    qrshow = mock_qr.show()
    assert {'Start Time', 'End Time', 'Instrument', 'Physobs', 'Source', 'Provider', 'Level', 'Wavelength', 'url'}.issubset(set(qrshow.colnames))



from unittest.mock import patch

import pytest
from parfive import Results
from parfive.results import Error

from sunpy.data.data_manager.downloader import DownloaderError, ParfiveDownloader


def test_ParfiveDownloader_errors():
    """
    Test that ParfiveDownloader raises an error when the download fails.
    """
    downloader = ParfiveDownloader()
    results = Results()
    results.errors.append(Error("", "FAKE_URL", ValueError("TEST_ERROR")))
    with patch('parfive.Downloader.download') as download:
        download.return_value = results
        with pytest.raises(DownloaderError, match='TEST_ERROR'):
            downloader.download('https://www.fakewebsite.com', 'test_file')



import numpy as np
import pytest

import astropy.units as u
from astropy.coordinates import Longitude
from astropy.tests.helper import assert_quantity_allclose

from sunpy.sun.models import differential_rotation


@pytest.fixture
def seconds_per_day():
    return 24 * 60 * 60.0 * u.s


def test_single(seconds_per_day):
    rot = differential_rotation(10 * seconds_per_day, 30 * u.deg)
    assert_quantity_allclose(rot, 136.8216 * u.deg, rtol=1e-3)


def test_array(seconds_per_day):
    rot = differential_rotation(10 * seconds_per_day, np.linspace(-70, 70, 2) * u.deg)
    assert_quantity_allclose(rot, Longitude(np.array([110.2725, 110.2725]) * u.deg), rtol=1e-3)


def test_synodic(seconds_per_day):
    rot = differential_rotation(10 * seconds_per_day, 30 * u.deg, model='howard', frame_time='synodic')
    assert_quantity_allclose(rot, 126.9656 * u.deg, rtol=1e-3)


def test_sidereal(seconds_per_day):
    rot = differential_rotation(10 * seconds_per_day, 30 * u.deg, model='howard', frame_time='sidereal')
    assert_quantity_allclose(rot, 136.8216 * u.deg, rtol=1e-3)


def test_howard(seconds_per_day):
    rot = differential_rotation(10 * seconds_per_day, 30 * u.deg, model='howard')
    assert_quantity_allclose(rot, 136.8216 * u.deg, rtol=1e-3)


def test_allen(seconds_per_day):
    rot = differential_rotation(10 * seconds_per_day, 30 * u.deg, model='allen')
    assert_quantity_allclose(rot, 136.9 * u.deg, rtol=1e-3)


def test_snodgrass(seconds_per_day):
    rot = differential_rotation(10 * seconds_per_day, 30 * u.deg, model='snodgrass')
    assert_quantity_allclose(rot, 135.4232 * u.deg, rtol=1e-3)


def test_rigid(seconds_per_day):
    rot = differential_rotation(10 * seconds_per_day, [0, 30, 60] * u.deg, model='rigid')
    assert_quantity_allclose(rot, [141.844 * u.deg] * 3, rtol=1e-3)


def test_fail(seconds_per_day):
    with pytest.raises(ValueError, match="model must equal one of { howard , snodgrass , allen , rigid }"):
        differential_rotation(10 * seconds_per_day, 30 * u.deg, model='garbage')



"""
Test cases for IRIS SJIMap
"""
import pytest

import astropy.units as u

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.sources.iris import SJIMap
from sunpy.util.exceptions import SunpyMetadataWarning
from .helpers import _test_private_date_setters


@pytest.fixture
def irismap():
    header = get_test_filepath("iris_l2_20130801_074720_4040000014_SJI_1400_t000.header")
    return get_dummy_map_from_header(header)


def test_fitstoIRIS(irismap):
    """Tests the creation of SJIMap using FITS."""
    assert (isinstance(irismap, SJIMap))


def test_private_date_setters(irismap):
    _test_private_date_setters(irismap)


def test_is_datasource_for(irismap):
    """Test the is_datasource_for method of SJIMap.
    Note that header data to be provided as an argument
    can be a MetaDict object."""
    assert irismap.is_datasource_for(irismap.data, irismap.meta)


def test_observatory(irismap):
    """Tests the observatory property of SJIMap."""
    assert irismap.observatory == "IRIS"


def test_wavelength(irismap):
    """Tests the wavelength and waveunit property of the SJIMap"""
    assert irismap.wavelength == u.Quantity(1400, 'Angstrom')


def test_level_number(irismap):
    """Tests the processing_level property of the SJIMap"""
    assert irismap.processing_level == 2.0


def test_units(irismap):
    """Tests the unit property of the SJIMap"""
    assert irismap.unit == u.Unit("DN")


def test_wcs(irismap):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    with pytest.warns(SunpyMetadataWarning, match='Missing metadata for observer'):
        irismap.pixel_to_world(0*u.pix, 0*u.pix)



"""
Access the Helio Event Catalogue
"""
import io
import os

from lxml import etree
from requests import Session
from requests.exceptions import SSLError
from zeep import Client
from zeep.transports import Transport

from astropy.io.votable.table import parse_single_table

from sunpy import log
from sunpy.net import attrs as a
from sunpy.net.base_client import BaseClient, QueryResponseTable
from sunpy.net.helio import attrs as ha
from sunpy.net.helio import parser
from sunpy.time import parse_time
from sunpy.util.exceptions import warn_user

__all__ = ['HECClient', 'HECResponse']


def votable_handler(xml_table):
    """
    Returns a VOtable object from a VOtable style xml string

    In order to get a VOtable object, it has to be parsed from an xml file or
    file-like object. This function creates a file-like object via the
    StringIO module, writes the xml data to it, then passes the file-like
    object to parse_single_table() from the astropy.io.votable.table module
    and thereby creates a VOtable object.

    Parameters
    ----------
    xml_table : `bytes`
        Contains the VOtable style xml data

    Returns
    -------
    votable : `astropy.io.votable.tree.Table`
        A properly formatted VOtable object

    """
    fake_file = io.BytesIO()
    fake_file.write(xml_table)
    votable = parse_single_table(fake_file)
    for i in range(len(votable.array)):
        item = votable.array[i][0]
        if isinstance(item, bytes):
            votable.array[i] = (votable.array[i][0].decode(),)
    fake_file.close()
    return votable


class HECResponse(QueryResponseTable):
    """
    A container for data returned from HEC searches.
    """


class HECClient(BaseClient):
    """
    Provides access to the HELIO webservices.
    """

    def __init__(self, link=None):
        """
        The constructor; establishes the webservice link for the client

        Initializes the client with a weblink

        Parameters
        ----------
        link : str
            Contains URL to valid WSDL endpoint

        Examples
        --------
        >>> from sunpy.net.helio import hec
        >>> hc = hec.HECClient()  # doctest: +SKIP
        """
        if link is None:
            # The default wsdl file
            link = parser.wsdl_retriever()
        session = Session()
        # This is for use in our test suite.
        session.verify = not (bool(os.environ.get("NO_VERIFY_HELIO_SSL", 0)))
        transport = Transport(session=session)
        try:
            self.hec_client = Client(link, transport=transport)
        except SSLError:
            log.warning('SSL verification error for HEC client.\n'
                     'Set the \'NO_VERIFY_HELIO_SSL\' environment variable disable SSL verification for Helio.')
            self.hec_client = None

    @classmethod
    def _can_handle_query(cls, *query):
        required = {a.Time}
        optional = {ha.MaxRecords, ha.TableName}
        return cls.check_attr_types_in_query(query, required, optional)

    @classmethod
    def _attrs_module(cls):
        return 'helio', 'sunpy.net.helio.attrs'

    def search(self, *args, **kwargs):
        """
        The simple interface to query the wsdl service.

        Used to utilize the service's TimeQuery() method, this is a simple
        interface between the sunpy module library and the web-service's API.

        .. note::
           By default the maximum records returned by the service are limited to 500.
           To obtain more results ``a.helio.MaxRecords`` must be set to a higher value.

        Examples
        --------
        >>> from sunpy.net.helio import attrs as ha
        >>> from sunpy.net import attrs as a, Fido
        >>> timerange = a.Time('2005/01/03', '2005/12/03')
        >>> res = Fido.search(timerange, ha.MaxRecords(10),
        ...                   ha.TableName('rhessi_hxr_flare'))  # doctest: +SKIP
        >>> res   # doctest: +SKIP
        <sunpy.net.fido_factory.UnifiedResponse object at ...>
        Results from 1 Provider:
        <BLANKLINE>
        10 Results from the HECClient:
        hec_id      time_start          time_peak      ... energy_kev flare_number
        ------ ------------------- ------------------- ... ---------- ------------
         31463 2005-01-03T01:37:36 2005-01-03T01:37:54 ...          6      5010320
         31464 2005-01-03T01:51:36 2005-01-03T01:59:18 ...         12      5010301
         31465 2005-01-03T03:26:28 2005-01-03T03:42:50 ...          6      5010332
         31466 2005-01-03T03:46:04 2005-01-03T04:07:10 ...         12      5010302
         31467 2005-01-03T05:00:24 2005-01-03T05:00:30 ...          6      5010313
         31468 2005-01-03T06:40:48 2005-01-03T06:42:46 ...          6      5010314
         31469 2005-01-03T08:27:56 2005-01-03T08:28:26 ...          6      5010334
         31470 2005-01-03T09:31:00 2005-01-03T09:33:34 ...          6      5010322
         31471 2005-01-03T09:34:52 2005-01-03T09:59:46 ...          6      5010336
         31472 2005-01-03T11:06:48 2005-01-03T11:07:18 ...         12      5010304
        <BLANKLINE>
        <BLANKLINE>
        """
        if self.hec_client is None:
            table = HECResponse([], client=self)
            return table
        qrdict = {}
        for elem in args:
            if isinstance(elem, a.Time):
                qrdict['Time'] = elem
            elif isinstance(elem, ha.MaxRecords):
                qrdict['max_records'] = elem.value
            elif isinstance(elem, ha.TableName):
                qrdict['table_name'] = elem.value
            else:
                raise ValueError(
                    f"{elem.__class__.__name__} should be a ``attrs.Time``, ``attrs.hek.MaxRecords`` or ``attrs.hek.TableName`` attribute.")
        qrdict.update(kwargs)
        table = qrdict.get('table_name', None)
        start_time = qrdict['Time'].start
        end_time = qrdict['Time'].end
        max_records = qrdict.get('max_records', 500)
        while table is None:
            table = self.select_table()
        start_time = parse_time(start_time)
        end_time = parse_time(end_time)
        results = self.hec_client.service.TimeQuery(STARTTIME=start_time.isot,
                                                    ENDTIME=end_time.isot,
                                                    FROM=table,
                                                    MAXRECORDS=max_records)
        results = votable_handler(etree.tostring(results))
        table = HECResponse(results.to_table(), client=self)
        if len(table) == max_records:
            warn_user("Number of results is the same as current limit. "
                      "It is possible your query has been truncated. "
                      "If you want to change this, set `a.helio.MaxRecords` to a higher value.")
        return table

    def get_table_names(self):
        """
        Returns a list of the available tables to query.

        Returns the names of all the tables that can be queried via the
        webservice.

        Returns
        -------
        tables.array: `numpy.ma.core.MaskedArray`
            A VOtable table of available tables names.

        Examples
        --------
        >>> from sunpy.net.helio import hec
        >>> hc = hec.HECClient()  # doctest: +SKIP
        >>> print(hc.get_table_names())  # doctest: +SKIP
        [('timed_see_flare',) ('hi_event',) ('yohkoh_flare_list',)
         ('wind_mfi_bs_crossing_time',) ('seeds_soho',) ('seeds_stb',)
         ...
         ('rhessi_hxr_flare',) ('cactus_soho_flow',) ('cactus_soho_cme',)
         ('stereob_het_sep',)]
        """
        results = self.hec_client.service.getTableNames()
        tables = votable_handler(etree.tostring(results))
        return tables.array

    def select_table(self):
        """
        Creates a list of table names and prompts the user for a choice

        This takes the table of table names from get_table_names(), creates a
        list of the names, sorts them, then presents the tables in a
        convenient menu for the user to choose from. It returns a string
        containing the name of the table that the user picked.

        Returns
        -------
        `str`
            Contains the name of the table that the user picked.

        Examples
        --------
        >>> from sunpy.net.helio import hec  # doctest: +SKIP
        >>> hc = hec.HECClient()  # doctest: +SKIP
        >>> hc.select_table()  # doctest: +SKIP
        """
        tables = self.get_table_names()
        table_list = sorted([t[0] for t in tables if len(t[0]) > 0])
        for index, table in enumerate(table_list):
            print(f'{index + 1} - {table}')
        while True:
            user_input = input(f"\nPlease enter a table number between 1 and {len(table_list)} "
                               "('e' to exit): ")
            if user_input.lower() == "e" or user_input.lower() == "exit":
                return None
            if user_input.isdigit() and 1 <= int(user_input) <= len(table_list):
                table_no = int(user_input)
                return table_list[table_no - 1]
            else:
                print(f"Input must be an integer between 1 and {len(table_list)}")

    def fetch(self, *args, **kwargs):
        """
        This is a no operation function as this client does not download data.
        """
        return NotImplemented



import os
import sys
from functools import wraps

import aiohttp
import parfive
from packaging.version import Version
from parfive import Results, SessionConfig

import sunpy

__all__ = ["Downloader", "Results"]

parfive_version = Version(parfive.__version__)
sunpy_headers = {
    "User-Agent": f"sunpy/{sunpy.__version__} parfive/{parfive.__version__} "
    f"aiohttp/{aiohttp.__version__} python/{sys.version[:5]} "
}
if os.environ.get("SUNPY_PYTEST_RUN"):
    # We want to allow remote servers to filter users from our online tests
    sunpy_headers["User-Agent"] = f"sunpy-CI-tests/{sunpy.__version__} parfive/{parfive.__version__} "
config = SessionConfig(headers=sunpy_headers)


class Downloader(parfive.Downloader):
    @wraps(parfive.Downloader.__init__)
    def __init__(self, *args, **kwargs):
        if "config" not in kwargs:
            kwargs["config"] = config
        super().__init__(*args, **kwargs)



from matplotlib.figure import Figure

import astropy.units as u
from astropy.wcs import WCS

from sunpy.tests.helpers import figure_test
from sunpy.visualization import show_hpr_impact_angle


@figure_test
def test_show_hpr_impact_angle():
    wcs = WCS({
        'ctype1': 'HRLN-CAR',
        'ctype2': 'HRLT-CAR',
        'cdelt1': 1,
        'cdelt2': 1,
        'crpix1': 180.5,
        'crpix2': 90.5,
        'crval1': 180,
        'crval2': 0,
    })

    fig = Figure(figsize=(9, 4))
    ax1 = fig.add_subplot(121, projection=wcs)
    ax1.grid()

    ax1.set_xlim(-0.5, 360 - 0.5)
    ax1.set_ylim(-0.5, 180 - 0.5)
    ax1.coords[0].set_ticks(spacing=90*u.deg)
    ax1.coords[1].set_ticks(spacing=30*u.deg)
    ax1.set_aspect(2)

    ax1.set_title("Shows declination by default")

    ax2 = fig.add_subplot(122, projection=wcs)
    ax2.grid()

    show_hpr_impact_angle(ax2.coords[1])

    ax2.set_xlim(-0.5, 360 - 0.5)
    ax2.set_ylim(-0.5, 180 - 0.5)
    ax2.coords[0].set_ticks(spacing=90*u.deg)
    ax2.coords[1].set_ticks(spacing=30*u.deg)
    ax2.set_aspect(2)

    ax2.set_title("Changed to show impact angle")

    return fig



"""
This file contains the entry points for asdf.
"""

import importlib.resources as importlib_resources

from asdf.extension import ManifestExtension
from asdf.resource import DirectoryResourceMapping


def get_resource_mappings():
    """
    Get the resource mapping instances for myschemas
    and manifests. This method is registered with the
    asdf.resource_mappings entry point.

    Returns
    -------
    list of collections.abc.Mapping
    """
    from . import resources
    resources_root = importlib_resources.files(resources)

    return [
        DirectoryResourceMapping(
            resources_root / "schemas", "asdf://sunpy.org/sunpy/schemas/"),
        DirectoryResourceMapping(
            resources_root / "manifests", "asdf://sunpy.org/sunpy/manifests/"),
    ]


def get_extensions():
    """
    Get the list of extensions.
    """
    from sunpy.io.special.asdf.converters.frames import SUNPY_FRAME_CONVERTERS
    from sunpy.io.special.asdf.converters.generic_map import GenericMapConverter

    sunpy_converters = [GenericMapConverter()] + SUNPY_FRAME_CONVERTERS
    # order here is important as asdf will pick the first matching
    # extension, start with the newest
    _manifest_uris = [
        "asdf://sunpy.org/sunpy/manifests/sunpy-1.2.1",
        "asdf://sunpy.org/sunpy/manifests/sunpy-1.2.0",
        "asdf://sunpy.org/sunpy/manifests/sunpy-1.1.2",
        "asdf://sunpy.org/sunpy/manifests/sunpy-1.1.1",
        "asdf://sunpy.org/sunpy/manifests/sunpy-1.1.0",
        "asdf://sunpy.org/sunpy/manifests/sunpy-1.0.0",
    ]

    return [
        ManifestExtension.from_uri(uri,
                                   converters=sunpy_converters,
                                   # Register that this is a replacement for
                                   # the old extension so old files still work.
                                   # without throwing a warning.
                                   legacy_class_names=["sunpy.io.special.asdf.extension.SunpyExtension"])
        for uri
        in _manifest_uris
    ]



"""
This module provides the base registration factory used for all SunPy data/net
factories.
"""
import inspect

__all__ = ["BasicRegistrationFactory", "NoMatchError",
           "MultipleMatchError", "ValidationFunctionError"]


class BasicRegistrationFactory:
    """
    Generalized registerable factory type.

    Widgets (classes) can be registered with an instance of this class.
    Arguments to the factory's ``__call__`` method are then passed to a function
    specified by the registered factory, which validates the input and returns
    a instance of the class that best matches the inputs.

    Attributes
    ----------
    registry : `dict`
        Dictionary mapping classes (key) to function (value) which validates input.
    default_widget_type : `type`
        Class of the default widget.
    validation_functions : `list` of `str`
        List of function names that are valid validation functions.

    Parameters
    ----------
    default_widget_type : `type`, optional
        Class of the default widget. Defaults to `None`.
    additional_validation_functions : `list` of `str`, optional
        List of strings corresponding to additional validation function names.
        Defaults to `list`.
    registry : `dict`, optional
        Dictionary mapping classes (key) to function (value) which validates input.
        Defaults to `None`.

    Notes
    -----
    * A valid validation function must be a classmethod of the registered widget
      and it must return a `bool`.
    """

    def __init__(self, default_widget_type=None,
                 additional_validation_functions=[], registry=None):
        if registry is None:
            self.registry = dict()
        else:
            self.registry = registry

        self.default_widget_type = default_widget_type

        self.validation_functions = (['_factory_validation_function'] +
                                     additional_validation_functions)

    def __call__(self, *args, **kwargs):
        """
        Method for running the factory.

        Arguments args and kwargs are passed through to the validation
        function and to the constructor for the final type.
        """
        # Any preprocessing and massaging of inputs can happen here

        return self._check_registered_widget(*args, **kwargs)

    def _check_registered_widget(self, *args, **kwargs):
        """
        Implementation of a basic check to see if arguments match a widget.
        """
        candidate_widget_types = list()

        for key in self.registry:

            # Call the registered validation function for each registered class
            if self.registry[key](*args, **kwargs):
                candidate_widget_types.append(key)

        n_matches = len(candidate_widget_types)

        if n_matches == 0:
            if self.default_widget_type is None:
                raise NoMatchError("No types match specified arguments and no default is set.")
            else:
                candidate_widget_types = [self.default_widget_type]
        elif n_matches > 1:
            raise MultipleMatchError(f"Too many candidate types identified ({n_matches})."
                                     "Specify enough keywords to guarantee unique type "
                                     "identification.")

        # Only one is found
        WidgetType = candidate_widget_types[0]

        return WidgetType(*args, **kwargs)

    def register(self, WidgetType, validation_function=None, is_default=False):
        """
        Register a widget with the factory.

        If ``validation_function`` is not specified, tests ``WidgetType`` for
        existence of any function in in the list ``self.validation_functions``,
        which is a list of strings which must be callable class attribute.

        Parameters
        ----------
        WidgetType : `type`
            Widget to register.
        validation_function : `function`, optional
            Function to validate against. Defaults to None, which indicates
            that a classmethod in validation_functions is used. Defaults to `None`.
        is_default : `bool`, optional
            Sets WidgetType to be the default widget. Defaults to `False`.
        """
        if is_default:
            self.default_widget_type = WidgetType
        elif validation_function is not None:
            if not callable(validation_function):
                raise AttributeError("Keyword argument 'validation_function' must be callable.")
            self.registry[WidgetType] = validation_function
        else:
            found = False
            for vfunc_str in self.validation_functions:
                if hasattr(WidgetType, vfunc_str):
                    vfunc = getattr(WidgetType, vfunc_str)
                    # check if classmethod: stackoverflow #19227724
                    _classmethod = inspect.ismethod(vfunc) and vfunc.__self__ is WidgetType
                    if _classmethod:
                        self.registry[WidgetType] = vfunc
                        found = True
                        break
                    else:
                        raise ValidationFunctionError(f"{WidgetType.__name__}.{vfunc_str} must be a classmethod.")
            if not found:
                raise ValidationFunctionError(f"No proper validation function for class {WidgetType.__name__} "
                                              "found.")

    def unregister(self, WidgetType):
        """
        Remove a widget from the factory's registry.
        """
        self.registry.pop(WidgetType)


class NoMatchError(Exception):
    """
    Exception for when no candidate class is found.
    """


class MultipleMatchError(Exception):
    """
    Exception for when too many candidate classes are found.
    """


class ValidationFunctionError(AttributeError):
    """
    Exception for when no candidate class is found.
    """



"""
===========================================
Drawing the AIA limb on a STEREO EUVI image
===========================================

In this example we use a STEREO-B and an SDO image to demonstrate how to
overplot the limb as seen by AIA on an EUVI-B image. Then we overplot the AIA
coordinate grid on the STEREO image.
"""
import matplotlib.pyplot as plt

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.coordinates.wcs_utils
import sunpy.map
from sunpy.data.sample import AIA_193_JUN2012, STEREO_B_195_JUN2012

##############################################################################
# Let's create a dictionary with the two maps, which we crop to full disk.

maps = {m.detector: m.submap(SkyCoord([-1100, 1100], [-1100, 1100],
                                      unit=u.arcsec, frame=m.coordinate_frame))
        for m in sunpy.map.Map([AIA_193_JUN2012, STEREO_B_195_JUN2012])}
maps['AIA'].plot_settings['vmin'] = 0  # set the minimum plotted pixel value

##############################################################################
# Now, let's plot both maps, and we draw the limb as seen by AIA onto the
# EUVI image. We remove the part of the limb that is hidden because it is on
# the far side of the Sun from STEREO's point of view.

fig = plt.figure(figsize=(10, 4))
ax1 = fig.add_subplot(121, projection=maps['AIA'])
maps['AIA'].plot(axes=ax1)
maps['AIA'].draw_limb(axes=ax1)

ax2 = fig.add_subplot(122, projection=maps['EUVI'])
maps['EUVI'].plot(axes=ax2)
visible, hidden = maps['AIA'].draw_limb(axes=ax2)
hidden.remove()

##############################################################################
# Let's plot the helioprojective coordinate grid as seen by SDO on the STEREO
# image in a cropped view. Note that only those grid lines that intersect the
# edge of the plot will have corresponding ticks and tick labels.

fig = plt.figure()
ax = fig.add_subplot(projection=maps['EUVI'])

maps['EUVI'].plot(axes=ax)

# Crop the view using pixel coordinates
ax.set_xlim(500, 1300)
ax.set_ylim(100, 900)

# Shrink the plot slightly and move the title up to make room for new labels.
ax.set_position([0.1, 0.1, 0.8, 0.7])
ax.set_title(ax.get_title(), pad=45)

# Change the default grid labels and line properties.
stereo_x, stereo_y = ax.coords
stereo_x.set_axislabel("Helioprojective Longitude (STEREO B) [arcsec]")
stereo_y.set_axislabel("Helioprojective Latitude (STEREO B) [arcsec]")
ax.coords.grid(color='white', linewidth=1)

# Add a new coordinate overlay in the SDO frame.
overlay = ax.get_coords_overlay(maps['AIA'].coordinate_frame)
overlay.grid()

# Configure the grid:
x, y = overlay

# Wrap the longitude at 180 deg rather than the default 360.
x.set_coord_type('longitude', 180*u.deg)

# Set the tick spacing
x.set_ticks(spacing=250*u.arcsec)
y.set_ticks(spacing=250*u.arcsec)

# Set the ticks to be on the top and left axes.
x.set_ticks_position('tr')
y.set_ticks_position('tr')

# Change the defaults to arcseconds
x.set_major_formatter('s.s')
y.set_major_formatter('s.s')

# Add axes labels
x.set_axislabel("Helioprojective Longitude (SDO) [arcsec]")
y.set_axislabel("Helioprojective Latitude (SDO) [arcsec]")

plt.show()



"""
Test cases for SUVI Map subclass.
"""
import pytest

import astropy.units as u

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.sources.suvi import SUVIMap
from .helpers import _test_private_date_setters


@pytest.fixture()
def suvi():
    """Creates an SUVIMap from a FITS file."""
    path = get_test_filepath("dr_suvi-l2-ci195_g16_s20190403T093200Z_e20190403T093600Z_v1-0-0_rebinned.header")
    return get_dummy_map_from_header(path)


def test_suvimap_creation(suvi):
    """Tests the creation of SUVIMap using FITS."""
    assert isinstance(suvi, SUVIMap)


def test_reference_date(suvi):
    assert suvi.reference_date.isot == "2019-04-03T09:32:33.340"


def test_date(suvi):
    assert suvi.date.isot == "2019-04-03T09:32:33.340"


def test_private_date_setters(suvi):
    _test_private_date_setters(suvi)


def test_is_datasource_for(suvi):
    """Test the is_datasource_for method of SUVIMap.
    Note that header data to be provided as an argument
    can be a MetaDict object."""
    assert suvi.is_datasource_for(suvi.data, suvi.meta)


def test_observatory(suvi):
    """Tests the observatory property of the SUVIMap object."""
    assert suvi.observatory == "GOES-R"


def test_detector(suvi):
    """Tests the detector property of the SUVIMap object."""
    assert suvi.detector == "SUVI"


def test_norm_clip(suvi):
    # Tests that the default normalizer has clipping disabled
    assert not suvi.plot_settings['norm'].clip


# SUVI provides observer coordinate information in an OBSGEO system, so this test
# needs remote data to access the latest IERS table to do a coordinate transformation from
# OBSGEO to heliographic Stonyhurst coordinates.
@pytest.mark.remote_data
def test_wcs(suvi):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    suvi.pixel_to_world(0*u.pix, 0*u.pix)



"""
================================
Combining off-limb and disk maps
================================

This example combines creating a composite plot with masking out the
solar disk. The purpose is to introduce users to the plotting process
required to overlay multiple maps that have been modified with other
``sunpy`` functions. The resulting plot in this tutorial shows information
on the upper photosphere, quiet corona, and magnetogram contours.
"""
import matplotlib.pyplot as plt

import astropy.units as u

import sunpy.data.sample
from sunpy.map import Map
from sunpy.map.maputils import all_coordinates_from_map, coordinate_is_on_solar_disk

###############################################################################
# Let's import sample data representing the three types of data we want to
# overlay: AIA 1600 (upper photosphere), AIA 171 (quiet corona), and HMI
# data (magnetic fields in the photosphere).

aia_1600 = Map(sunpy.data.sample.AIA_1600_IMAGE)
aia_171 = Map(sunpy.data.sample.AIA_171_IMAGE)
hmi = Map(sunpy.data.sample.HMI_LOS_IMAGE)

###############################################################################
# Next, let's mask out the solar disk of the AIA 171 image since we will be
# overlaying the AIA 1600 photosphere.

hpc_coords = all_coordinates_from_map(aia_171)
mask = coordinate_is_on_solar_disk(hpc_coords)
colormap = aia_171.cmap.copy()
colormap.set_bad('black')

scaled_map = Map(aia_171.data, aia_171.meta, mask=mask)

###############################################################################
# Before we plot the composite image, let's define the contour levels that will
# be applied to the HMI image. Also, grab the AIA 171 colormap to use with
# the AIA 1600 image to better match the images.

levels = [-1000, -500, -250, 250, 500, 1000]*u.G
aia171_cmap = plt.get_cmap('sdoaia171')

###############################################################################
# Now, create the plot by overlaying the maps and contours. Note that
# ``clip_interval`` is being defined to highlight the coronal features of the
# scaled AIA 171 map and the AIA 1600 map is autoaligned to the AIA 171 WCS
# Axes.

fig = plt.figure()
ax = fig.add_subplot(projection=scaled_map)
scaled_map.plot(axes=ax, clip_interval=(1, 99.95)*u.percent,
                cmap=colormap, zorder=0)
aia_1600.plot(axes=ax, cmap=aia171_cmap, autoalign=True, zorder=1)
hmi.draw_contours(axes=ax, levels=levels, zorder=2)
ax.set_title("AIA 171 (off limb) + AIA 1600 (on disk) + HMI (contours)")

plt.show()



import copy

import numpy as np
import pytest

import astropy.table
import astropy.time
import astropy.units as u
import astropy.utils
import astropy.utils.masked
from astropy.coordinates import SkyCoord
from astropy.tests.helper import assert_quantity_allclose

from sunpy.coordinates import (
    Heliocentric,
    HeliographicCarrington,
    HeliographicStonyhurst,
    Helioprojective,
    get_earth,
)
from sunpy.net.hek.utils import (
    _map_chain_code_columns_to_coordinates,
    _map_columns_to_quantities,
    _map_event_coord_columns_to_coordinates,
    _parse_unit,
)


@pytest.mark.parametrize(('input_unit', 'expected_unit'), [
    ('DN/sec/pixel', u.DN/(u.pix*u.s)),
    ('ergs per cubic centimeter', u.erg/u.cm**3),
    ('HMI pixel', 0.5*u.arcsec/u.pixel),
    ('HMI pixels', 0.5*u.arcsec/u.pixel),
    ('ma/m2', u.milliampere/u.m**2),
    ('deg deg', u.deg),
    ('arcsec, arcsec', u.arcsec),
    ('arcsec,arcsec', u.arcsec),
    ("arcseconds", u.arcsec),
    ("degrees", u.deg),
    ("sec", u.s),
    ("emx", u.Mx),
    ("amperes", u.A),
    ("ergs", u.erg),
    ("radians", u.rad),
    ("hz", u.Hertz),
    ("rsun", u.R_sun),
    ("cubic centimeter", u.cm**3),
    ("square centimeter", u.cm**2),
    ("cubic meter", u.m**3),
    ("square meter", u.m**2),
    ("ergs/cm^3", u.erg/u.cm**3),
])
def test_parse_unit(input_unit, expected_unit):
    assert _parse_unit(input_unit) == expected_unit


def test_parse_unit_invalid():
    """
    Test that an unexpected unit errors in the same way as u.Unit
    """
with pytest.raises(ValueError, match=r"'foo' did not parse as unit"):
    _parse_unit('foo')


def map_event_coords_to_column(coord):
    """
    Given a coordinate, create a table that mimics what is returned by the HEK. This is to test
    roundtripping of coordinates through the logic used to parse the "event_coord" columns.
    """
    frame_sys_mapping = {
        "heliographic_stonyhurst": "UTC-HGS-TOPO",
        "helioprojective": "UTC-HPC-TOPO",
        "heliographic_carrington": "UTC-HGC-TOPO",
        "heliocentric": "UTC-HRC-TOPO",
    }
    columns = {
        "event_coordsys": [frame_sys_mapping[coord.frame.name]],
        "event_coordunit": [coord.info.unit],
        "event_starttime": [coord.obstime.iso],
    }
    for i, c in enumerate(coord.data.components):
        columns[f"event_coord{i+1}"] = [getattr(coord.data, c).value]
    # Swap order of first two coords and delete third in the case of HCC (radial)
    if coord.frame.name == "heliocentric":
        columns["event_coord1"], columns["event_coord2"] = columns["event_coord2"], columns["event_coord1"]
        columns["event_coord1"] = [columns["event_coord1"][0] - 90]
        del columns["event_coord3"]
        columns['event_coordunit'] = [', '.join(columns["event_coordunit"][0].split(",")[:-1][::-1])]
    return astropy.table.Table(columns)


def test_event_coord_column_mapping():
    obstime = "2020-01-01"
    observer = get_earth(obstime)
    input_coordinates = [
        SkyCoord(100,200, unit='arcsec', frame=Helioprojective(observer=observer)),
        SkyCoord(20, 45, unit='degree', frame=HeliographicStonyhurst(obstime=obstime)),
        SkyCoord(20*u.deg, 45*u.deg, 1*u.R_sun, frame=HeliographicStonyhurst(obstime=obstime)),
        SkyCoord(180*u.deg, 0*u.deg, frame=HeliographicCarrington(observer=observer)),
        SkyCoord(0.1, 45, 1, unit=['R_sun', 'deg', 'R_sun'], frame=Heliocentric(observer=observer), representation_type='cylindrical'),
    ]
    hek_table = astropy.table.vstack([map_event_coords_to_column(ic) for ic in input_coordinates])
    _map_event_coord_columns_to_coordinates(hek_table)
    assert "event_coordunit" in hek_table.colnames
    for ic, hc in zip(input_coordinates, hek_table['event_coord']):
        assert ic == hc


def test_quantity_column_mapping():
    data = {
        "meanvertcurrentdensity": [1.0, None, 3.0],
        "currentdensityunit": ["ma/m2", None, "ma/m2"],
        "meanphotoenergydensity": [20.0, None, 30.0],
        "meanenergydensityunit": ["ergs per cubic centimeter", None, "ergs per cubic centimeter"],
        "boundbox_c1ur": [200, 250, 0],
        "event_coordunit": ["arcsec, arcsec", "arcsec", "deg deg"],
    }
    hek_table = astropy.table.Table(data)
    _map_columns_to_quantities(hek_table)
    assert u.allclose(hek_table["meanvertcurrentdensity"],
                      astropy.utils.masked.Masked(data["meanvertcurrentdensity"]*u.mA/u.m**2, mask=[False, True, False]))
    assert u.allclose(hek_table["meanphotoenergydensity"],
                      astropy.utils.masked.Masked(data["meanphotoenergydensity"]*u.erg/u.cm**3, mask=[False, True, False]))
    assert u.allclose(hek_table["boundbox_c1ur"], data["boundbox_c1ur"]*u.arcsec)
    assert hek_table.colnames == ["meanvertcurrentdensity", "meanphotoenergydensity", "boundbox_c1ur"]


obstime = astropy.time.Time("2010-01-01") + np.arange(5) * u.year
hpc_frame = Helioprojective(observer=get_earth(obstime))
hgs_frame = HeliographicStonyhurst(obstime=obstime)
hgc_frame = HeliographicCarrington(observer=get_earth(obstime))
hcc_frame = Heliocentric(observer=get_earth(obstime))


def get_point_chaincode(coord):
    if coord.frame.name == "heliocentric":
        components = coord.data.components[:-1]
    else:
        components = coord.data.components
    data = [copy.copy(getattr(coord.data, c).value) for c in components]
    if coord.frame.name == "heliocentric":
        data[-1] -= 90.0
    return f"POINT({' '.join([str(d) for d in data])})"


@pytest.mark.parametrize(("colname", "coord"), [
    ("hpc_coord", SkyCoord(*np.random.rand(2,len(obstime))*u.arcsec, frame=hpc_frame)),
    ("hgs_coord", SkyCoord(*np.random.rand(2,len(obstime))*u.deg, frame=hgs_frame)),
    ("hgc_coord", SkyCoord(*np.random.rand(2,len(obstime))*u.deg, frame=hgc_frame)),
    ("hrc_coord", SkyCoord(psi=np.random.rand(len(obstime))*360*u.deg,
                           rho=np.random.rand(len(obstime))*u.R_sun,
                           z=1*u.R_sun,
                           frame=hcc_frame,
                           representation_type='cylindrical')),
])
def test_chaincode_point_column_mapping(colname, coord):
    table = astropy.table.Table({
        "event_starttime": obstime.iso,
        colname: [get_point_chaincode(c) for c in coord],
    })
    _map_chain_code_columns_to_coordinates(table)
    assert isinstance(table[colname], SkyCoord)
    assert table[colname].frame.is_equivalent_frame(coord.frame)
    assert_quantity_allclose(table[colname].cartesian.xyz, coord.cartesian.xyz)


def get_polygon_chaincode(coord):
    if coord is None or coord == '':
        return coord
    c1 = copy.copy(getattr(coord.data, coord.data.components[0]).value)
    c2 = copy.copy(getattr(coord.data, coord.data.components[1]).value)
    if coord.frame.name == 'heliocentric':
        c2 -= 90.0
    data = [f'{c1} {c2}' for c1, c2 in zip(c1, c2)]
    return f"POLYGON(({','.join(data)}))"


@pytest.mark.parametrize(("colname", "coord"), [
    ("hpc_bbox", SkyCoord(*np.random.rand(2,7,len(obstime))*u.arcsec, frame=hpc_frame).T),
    ("hrc_bbox", SkyCoord(psi=np.random.rand(9,len(obstime))*360*u.deg,
                          rho=np.random.rand(9,len(obstime))*u.R_sun,
                          z=1*u.R_sun,
                          frame=hcc_frame,
                          representation_type='cylindrical').T),
    ("bound_chaincode", [SkyCoord(*np.random.rand(2,nb), unit='deg', frame='icrs') for nb in np.random.randint(6,11, size=len(obstime))]),
])
def test_chaincode_polygon_column_mapping(colname, coord):
    table = astropy.table.Table({
        "event_starttime": obstime.iso,
        "event_coordunit": len(obstime)*["deg"],
        colname: [get_polygon_chaincode(c) for c in coord],
    })
    _map_chain_code_columns_to_coordinates(table)
    if isinstance(table[colname], SkyCoord):
        assert table[colname].frame.is_equivalent_frame(coord.frame)
        assert_quantity_allclose(table[colname].cartesian.xyz, coord.cartesian.xyz)
    else:
        for i, c in enumerate(coord):
            assert isinstance(table[colname][i], SkyCoord)
            assert table[colname][i].frame.is_equivalent_frame(c.frame)
            assert_quantity_allclose(table[colname][i].cartesian.xyz, c.cartesian.xyz)



import pytest

from asdf.testing.helpers import roundtrip_object
from astropy.time import Time

# This registers TimeUTime in astropy.time.Time
from sunpy.time import TimeUTime  # NOQA

pytest.importorskip("asdf_astropy", "0.8.0")


def assert_roundtrip_time(old):
    new = roundtrip_object(old)
    assert new.utime == old.utime


def test_utime():
    assert_roundtrip_time(Time(1231853576.0, format='utime'))


def test_tai_seconds():
    assert_roundtrip_time(Time(1231853576.0, format='tai_seconds'))



import os
import json
import urllib.request
from pathlib import Path

import requests

from sunpy.net import attrs as a
from sunpy.net.attr import and_
from sunpy.net.base_client import BaseClient, QueryResponseTable
from sunpy.net.solarnet.attrs import Dataset, Tags, Target, walker

_BASE_URL = "https://solarnet2.oma.be/service/api/svo/{}"

__all__ = ["SOLARNETClient"]

class SOLARNETClient(BaseClient):
    """
    Provides access to query and download from the SOLARNET Virtual Observatory (SVO).

    The SVO was first supported by the `SOLARNET <http://solarnet-east.eu/>`__ project,
    funded by the European Commission's FP7 Capacities Programme under the Grant Agreement 312495.
    It was turned operational due to the `SOLARNET2 <https://solarnet-project.eu/>`__ project, funded by
    the European Union's Horizon 2020 Research and Innovation Programme under Grant Agreement 824135.
    It's purpose is to collect metadata from as many solar observations as possible,
    especially those involved in the SOLARNET projects, in a common catalog and make them available to
    the scientific community.

    There is various ways one can access SVO `Data <https://solarnet.oma.be/#introduction>`__, this
    client using the RESTful API (`available documentation <https://solarnet.oma.be/svo_restful_api_user_manual.html>`__).

    Notes
    -----
    This client by defaults limits the number of results returned from a search to 20.
    To change this limit, use the ``a.solarnet.Limit`` attribute in your query.

    In addition, the query returns a large number of columns, many of which are not useful to the user.
    These are columns found in the FITS headers of the files, and are hidden by default.
    If you want to see all the columns, you can do ``query_results.show()`` on the results
    returned from a search.

    Examples
    --------
    >>> from sunpy.net import Fido, attrs as a

    >>> search_results = Fido.search(a.solarnet.Dataset.gris_level_1,a.solarnet.Limit(2), a.solarnet.Target.ar)   # doctest: +REMOTE_DATA
    >>> search_results  # doctest: +REMOTE_DATA
    <sunpy.net.fido_factory.UnifiedResponse object at ...>
    Results from 1 Provider:
    <BLANKLINE>
    2 Results from the SOLARNETClient:
    Source: https://solarnet2.oma.be
    <BLANKLINE>
         OID               DATE_BEG                 DATE_END          WAVEMIN   WAVEMAX          DATE_OBS         EXPTIME MEASURE STEPS WAVELENG
    -------------- ------------------------ ------------------------ --------- --------- ------------------------ ------- ------- ----- --------
    20140426094659 2014-04-26T09:46:59.000Z 2014-04-26T09:53:41.300Z 1564.0199 1568.0461 2014-04-26T00:00:00.000Z    None          None     1566
    20140426095643 2014-04-26T09:56:43.000Z 2014-04-26T10:09:50.300Z 1564.0191 1568.0461 2014-04-26T00:00:00.000Z    None          None     1566
    <BLANKLINE>
    <BLANKLINE>
    """

    @property
    def info_url(self):
        return 'https://solarnet2.oma.be'

    def search(self, *query):
        """
        Searches the SVO for datasets matching the query.

        Parameters
        ----------
        *query : `~sunpy.net.attrs`
            Attributes to search for datasets.

        Returns
        -------
        `~sunpy.net.base_client.QueryResponseTable`
            A table containing the search results.
        """
        results = []
        query = and_(*query)
        block = walker.create(query)[0]
        if "datasets" in block:
            url = _BASE_URL.format(block["datasets"])
            block.pop("datasets")
        try:
            req = urllib.request.urlopen(url + '?' + urllib.parse.urlencode(block))
            data = json.loads(req.read().decode())["objects"]
        except Exception as e:
            raise OSError(f"Failed to fetch data from {url}") from e
        # It returns so many keys that are not useful for the user
        allowed_keys = [
            "oid", "date_beg", "date_end", "wavemin", "wavemax", "date_obs",
            "exptime", "measure", "steps", "telescope", "waveleng",
        ]
        hide_keys = [
            key.upper() for key in data[0].keys() if key.lower() not in allowed_keys
        ]
        # Just to avoid empty tags
        hide_keys.append("Tags") if "tag" not in block else hide_keys
        for i in data:
            # Filter out keys that should be hidden or not displayed
            filtered_data = {key.upper(): value for key, value in i.items()}
            # Add specific fields from data_location
            filtered_data.update({
                "F_SIZE": i["data_location"]["file_size"],
                "URL": i["data_location"]["file_url"],
            })
            results.append(filtered_data)
        qrt = QueryResponseTable(results, client=self)
        qrt.hide_keys = hide_keys + ["URL", "F_SIZE"]
        return qrt

    def fetch(self, query_results, path, downloader, **kwargs):
        for row in query_results:
            fname = row['URL'].split('/')[-1]
            filepath = str(path).format(file=fname)
            max_splits = kwargs.get('max_splits', 5)
            downloader.enqueue_file(row['URL'], filename=filepath, max_splits=max_splits)

    @staticmethod
    def load_solarnet_values():
        """
        Load Solarnet dataset values from a local JSON file.

        Returns
        -------
        attrs : `dict`
            Dictionary of Solarnet dataset values.
        """
        data_sets = Path(__file__).parent / "data" / "datasets.json"
        tags = Path(__file__).parent / "data" / "tags.json"
        with data_sets.open() as data_values:
            data = json.load(data_values)
        with tags.open() as tags_values:
            tags = json.load(tags_values)
        data_set = {Dataset: list(data.items())}
        target = {Target: [("AR", "Active Region"),("CH","Coronal Hole"),("FS","Flare"),("QR","Quiet Region")]}
        tags = {Tags:list(tags.items())}
        attrs = data_set | target | tags
        return attrs

    @staticmethod
    def _update_solarnet_data():
        """
        Gets the names of available datasets, tags, and targets from SolarNet and saves them into
        separate JSON files.
        """
        dir_path = os.path.dirname(os.path.realpath(__file__))
        data_dir = os.path.join(dir_path, 'data')
        os.makedirs(data_dir, exist_ok=True)
        # Datasets
        url_dataset = _BASE_URL.format("dataset")
        response = requests.get(url_dataset, params={"limit": 100})
        data = response.json()
        values = {}
        for obj in data.get("objects", []):
            name = obj["name"].replace(" ", "_").lower()
            description = obj.get("description") or obj.get("telescope", {}).get("description", "")
            values[name] = description.split(". ")[0].split(", ")[0] if description else ""
        with open(os.path.join(data_dir, 'datasets.json'), 'w') as f:
            json.dump(dict(sorted(values.items())), f, indent=2)
        # Tags
        url_tags = _BASE_URL.format("tag")
        response = requests.get(url_tags)
        data_tags = response.json()
        tag_values = {}
        for obj in data_tags.get("objects", []):
            tag_name = obj["name"].replace(" ", "_").lower()
            tag_values[tag_name] = tag_name
        with open(os.path.join(data_dir, 'tags.json'), 'w') as f:
            json.dump(dict(sorted(tag_values.items())), f, indent=2)

    @classmethod
    def register_values(cls):
        return cls.load_solarnet_values()

    @classmethod
    def _attrs_module(cls):
        return 'solarnet', 'sunpy.net.solarnet.attrs'

    @classmethod
    def _can_handle_query(cls, *query):
        return any(isinstance(q, a.solarnet.Dataset) for q in query)



import io
import os
from pathlib import Path
from contextlib import redirect_stdout

import pytest

import sunpy
from sunpy import config
from sunpy.extern.appdirs import AppDirs
from sunpy.util import SunpyUserWarning
from sunpy.util.config import (
    CONFIG_DIR,
    _find_config_files,
    _get_user_configdir,
    _is_writable_dir,
    copy_default_config,
    dirs,
    get_and_create_download_dir,
    get_and_create_sample_dir,
    print_config,
)

USER = os.path.expanduser('~')


def test_is_writable_dir(tmpdir, tmp_path):
    assert _is_writable_dir(tmpdir)
    tmp_file = tmpdir.join("hello.txt")
    # Have to write to the file otherwise its seen as a directory(?!)
    tmp_file.write("content")
    # Checks directory with a file
    assert _is_writable_dir(tmpdir)
    # Checks a filepath instead of directory
    assert not _is_writable_dir(tmp_file)


def test_get_user_configdir(tmpdir, tmp_path, undo_config_dir_patch):
    # Default
    assert USER in CONFIG_DIR
    assert CONFIG_DIR.split(os.path.sep)[-1] == "sunpy"
    assert CONFIG_DIR == _get_user_configdir()
    # Try to set a manual one (already created)
    tmp_config_dir = tmpdir.mkdir("config_here")
    os.environ["SUNPY_CONFIGDIR"] = tmp_config_dir.strpath
    assert tmp_config_dir == _get_user_configdir()
    # Bypass this under windows.
    if not (os.name == "nt"):
        os.unsetenv("SUNPY_CONFIGDIR")
    del os.environ["SUNPY_CONFIGDIR"]
    # Try to set a manual one (not created)
    os.environ["SUNPY_CONFIGDIR"] = str(tmp_path)
    assert str(tmp_path) == _get_user_configdir()
    # Bypass this under windows.
    if not (os.name == "nt"):
        os.unsetenv("SUNPY_CONFIGDIR")
    del os.environ["SUNPY_CONFIGDIR"]


def test_print_config_files(tmpdir, tmp_path, undo_download_dir_patch):
    with io.StringIO() as buf, redirect_stdout(buf):
        print_config()
        printed = buf.getvalue()
    assert "time_format = %Y-%m-%d %H:%M:%S" in printed
    assert _find_config_files()[0] in printed
    assert get_and_create_download_dir() in printed
    assert get_and_create_sample_dir() in printed


def test_find_config_user_site_files(tmpdir, tmp_path, undo_download_dir_patch, monkeypatch):

    # Set a site config dir
    tmp_config_site_dir = tmpdir.mkdir("sunpy_test_configdir_site")
    monkeypatch.setattr(AppDirs, "site_config_dir", tmp_config_site_dir.strpath)

    # Create sunpyrc to site configdir
    tmp_site_file = tmp_config_site_dir.join("sunpyrc")
    # Have to write to the file otherwise its seen as a directory(?!)
    tmp_site_file.write("content")

    # Test that the second config file of two is the site file
    config_files = _find_config_files()
    assert len(config_files) == 2
    assert config_files[1] == tmp_site_file.strpath

    # set a user config dir
    tmp_config_user_dir = tmpdir.mkdir("sunpy_test_configdir")
    monkeypatch.setenv("SUNPY_CONFIGDIR", tmp_config_user_dir.strpath)

    # Copy default sunpyrc to user configdir
    # With a sunpyrc file in the site config dir,
    # the function should raise a warning.
    with pytest.warns(SunpyUserWarning):
        copy_default_config()

    # Test that the second config file of three is the site file
    # and the third of three is the user config file
    config_files = _find_config_files()
    assert len(config_files) == 3
    assert config_files[1] == tmp_site_file.strpath
    assert config_files[2] == tmp_config_user_dir.join("sunpyrc").strpath


def test_get_and_create_download_dir(undo_download_dir_patch):
    # test default config
    path = get_and_create_download_dir()
    assert Path(path) == Path(USER) / 'sunpy' / 'data'
    # test updated config
    new_path = os.path.join(USER, 'sunpy_data_here_please')
    config.set('downloads', 'download_dir', new_path)
    path = get_and_create_download_dir()
    assert path == os.path.join(USER, new_path)
    # Set the config back
    os.rmdir(new_path)
    config.set('downloads', 'download_dir', os.path.join(USER, 'sunpy', 'data'))


def test_get_and_create_sample_dir():
    # test default config
    path = get_and_create_sample_dir()
    assert Path(path) == Path(dirs.user_data_dir)
    # test updated config
    new_path = os.path.join(USER, 'sample_data_here_please')
    config.set('downloads', 'sample_dir', new_path)
    path = get_and_create_sample_dir()
    assert path == new_path
    # Set the config back
    os.rmdir(new_path)
    config.set('downloads', 'sample_dir', os.path.join(USER, 'sunpy', 'data', 'sample_data'))


def test_copy_default_config(tmpdir, undo_config_dir_patch, monkeypatch):
    config_filename = 'sunpyrc'

    # Try to set a manual one (already created)
    tmp_config_dir = tmpdir.mkdir("sunpy_test_configdir")
    monkeypatch.setenv("SUNPY_CONFIGDIR", tmp_config_dir.strpath)

    assert tmp_config_dir == _get_user_configdir()

    config_file = Path(sunpy.__file__).parent / 'data' / config_filename
    user_config_file = tmp_config_dir / config_filename

    # Create a new config file
    copy_default_config()
    with open(user_config_file) as ucf:
        with open(config_file) as cf:
            assert ucf.read() == cf.read()


def test_copy_default_config_without_overwrite(tmpdir, undo_config_dir_patch, monkeypatch):

    # Try to set a manual one (already created)
    tmp_config_dir = tmpdir.mkdir("sunpy_test_configdir")
    monkeypatch.setenv("SUNPY_CONFIGDIR", tmp_config_dir.strpath)

    assert tmp_config_dir == _get_user_configdir()

    # Create a new config file
    copy_default_config()

    # Without the `overwrite` parameter
    # the function should raise an error.
    with pytest.warns(SunpyUserWarning):
        copy_default_config()


def test_copy_default_config_with_overwrite(tmpdir, undo_config_dir_patch, monkeypatch):

    # Try to set a manual one (already created)
    tmp_config_dir = tmpdir.mkdir("sunpy_test_configdir")
    monkeypatch.setenv("SUNPY_CONFIGDIR", tmp_config_dir.strpath)

    assert tmp_config_dir == _get_user_configdir()

    # Create a new config file
    copy_default_config()

    # With the `overwrite` parameter
    # the function does not raise any error.
    with pytest.warns(SunpyUserWarning):
        copy_default_config(overwrite=True)


def test_copy_default_config_sitewarn(tmpdir, undo_config_dir_patch, monkeypatch):

    # Set a user config dir
    tmp_config_dir = tmpdir.mkdir("sunpy_test_configdir")
    monkeypatch.setenv("SUNPY_CONFIGDIR", tmp_config_dir.strpath)

    assert tmp_config_dir == _get_user_configdir()

    # Set a site config dir
    tmp_config_site_dir = tmpdir.mkdir("sunpy_test_configdir_site")
    monkeypatch.setattr(AppDirs, "site_config_dir", tmp_config_site_dir.strpath)

    # Create sunpyrc to site configdir
    tmp_file = tmp_config_site_dir.join("sunpyrc")
    # Have to write to the file otherwise its seen as a directory(?!)
    tmp_file.write("content")

    # With a sunpyrc file in the site config dir,
    # the function should raise a warning.
    with pytest.warns(SunpyUserWarning):
        copy_default_config()



"""
=====================================================
Retrieving and analyzing GOES X-Ray Sensor (XRS) data
=====================================================

The X-ray Sensor (XRS) on board the GOES series of satellites
have provided soft X-ray measurements in two broadband energy
ranges 0.5-4 and 1-8 angstrom since 1975. The GOES 16 and 17
satellites are the latest in line. The flux levels in the GOES
1-8 angstrom channel are used to report flares and determine
their size (i.e. their GOES class).

In this example we are going to look at how you can query and
retrieve the GOES XRS data using `~sunpy.net.Fido` and load it
into a `~sunpy.timeseries.TimeSeries`.

Some things to note: NOAA have recently re-processed the GOES 13,
14 and 15 XRS science quality data, such that the SWPC scaling factor
has been removed. This means that the fluxes will have a different values,
and so will flare peak fluxes from previous 13, 14 and 15 XRS data. See
`here <https://www.ncei.noaa.gov/data/goes-space-environment-monitor/access/science/xrs/GOES_1-15_XRS_Science-Quality_Data_Readme.pdf>`__
for more details. The sunpy GOES XRS client for Fido now provides this
new re-processed data. We now also provide the data for GOES 16 and 17.

Another thing to note is that the GOES XRS client `~sunpy.net.Fido` now
returns all available GOES data for the specific timerange queried. For
example, there are times when GOES 13, 14 and 15 overlap and such data is
available from each satellite. Similarly there are times when GOES 16 and 17 overlap.
"""
import matplotlib.pyplot as plt
import numpy as np

from astropy.visualization import time_support

from sunpy import timeseries as ts
from sunpy.net import Fido
from sunpy.net import attrs as a

#############################################################
# Lets first define our start and end times and query using the
# `~sunpy.net.Fido`.

tstart = "2015-06-21 01:00"
tend = "2015-06-21 23:00"
result = Fido.search(a.Time(tstart, tend), a.Instrument("XRS"))
print(result)

#############################################################
# As we can see this now returns six results, two files for each GOES
# 13, one for GOES 14 and one for GOES 15, which can be identified
# by the ``SatelliteNumber`` column.
# The GOES data provided can either be high-cadence (1s/2s/3s based on GOES satellite) or averaged over 1 minute.
# This can be noted in the ``Resolution`` column, where the ``avg1m`` and ``flx1s`` attributes are the 1 minute average and the high-cadence data, respectively.
# However, we probably will only want one of these files for our analysis, so we can query by the `sunpy.net.attrs`:
# `sunpy.net.dataretriever.attrs.goes.SatelliteNumber` to specify what GOES satellite number we want
# to use, and `sunpy.net.attrs.Resolution` for the resolution. Here we will use the high-cadence observations.

result_goes15 = Fido.search(a.Time(tstart, tend), a.Instrument("XRS"), a.goes.SatelliteNumber(15), a.Resolution("flx1s"))
print(result_goes15)

#############################################################
# Now we can see that this returns just one file for the GOES 15 data.
# Lets now download this data using `~sunpy.net.fido_factory.UnifiedDownloaderFactory.fetch`.

file_goes15 = Fido.fetch(result_goes15)

#############################################################
# Lets now load this data into a `~sunpy.timeseries.TimeSeries`,
# and inspect the data using `~sunpy.timeseries.GenericTimeSeries.peek()`.

goes_15 = ts.TimeSeries(file_goes15)
goes_15.peek()

#############################################################
# The resulting `~sunpy.timeseries.TimeSeries` can be filtered by GOES quality flags. For more information
# refer to the `GOES Data Guide <https://www.ncei.noaa.gov/data/goes-space-environment-monitor/access/science/xrs/GOES_1-15_XRS_Science-Quality_Data_Readme.pdf>`__.

df = goes_15.to_dataframe()
df = df[(df["xrsa_quality"] == 0) & (df["xrsb_quality"] == 0)]
goes_15 = ts.TimeSeries(df, goes_15.meta, goes_15.units)

###############################################################
# We can also pull out the individual GOES channels and plot. The 0.5-4 angstrom
# channel is known as the "xrsa" channel and the 1-8 angstrom channel is known
# as the "xrsb" channel.

fig, ax = plt.subplots()
goes_15.plot(axes=ax, columns=["xrsb"])
plt.show()

###############################################################
# We can also truncate the data for the time of the large flare,
# and analyze the different channels. For example, we can plot the
# derivative which is useful in terms of the Neupert effect when analyzing
# flares.

goes_flare = goes_15.truncate("2015-06-21 09:35", "2015-06-21 10:30")

time_support()
fig, ax = plt.subplots(figsize=(10, 8))
ax.plot(goes_flare.time, np.gradient(goes_flare.quantity("xrsb")))
ax.set_ylabel("Flux (Wm$^{-2}$$s^{-1}$)")
fig.autofmt_xdate()
plt.show()

###############################################################
# GOES 16 and 17 data
# -------------------
# Since March 2020, data prior to GOES 15 (incl) is no longer supported
# by NOAA and GOES 16 and 17 data is now provided. See
# `here <https://data.ngdc.noaa.gov/platforms/solar-space-observing-satellites/goes/goes16/l2/docs/GOES-R_XRS_L2_Data_Users_Guide.pdf>`__
# for more details. GOES 16 and 17 are part of the GOES-R series and provide
# XRS data at a better time resolution (1s). sunpy now supports this data also.
# GOES 16 has been taking observations from 2017, and GOES 17 since 2018, both of
# which are now and its now available through sunpy.net.Fido.

###############################################################
# Lets query for some data over two days.

results = Fido.search(a.Time("2020-11-20 00:00", "2020-11-21 23:00"), a.Instrument("XRS"), a.Resolution("flx1s"))
print(results)

###############################################################
# We can see that we are provided with 4 results, two files for GOES 16
# and two for GOES 17. Again we can make the query only specifying one
# GOES satellite number.

results_16 = Fido.search(a.Time("2020-11-20 00:00", "2020-11-21 23:00"), a.Instrument("XRS"),
                         a.goes.SatelliteNumber(16))
print(results_16)

###############################################################
# Lets now download this data and load into a
# `~sunpy.timeseries.TimeSeries`.

files = Fido.fetch(results_16)
# We use the `concatenate=True` keyword argument in TimeSeries, as
# we have two files and want to create one timeseries from them.
goes_16 = ts.TimeSeries(files, concatenate=True)
goes_16.peek()



"""
Test the `GenericMap.reproject_to()` method
"""
import warnings

import numpy as np
import pytest
from matplotlib.figure import Figure
from matplotlib.testing.decorators import check_figures_equal

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.tests.helper import assert_quantity_allclose
from astropy.wcs import WCS

import sunpy.coordinates
import sunpy.map
from sunpy.data.test import get_dummy_map_from_header
from sunpy.tests.helpers import figure_test
from sunpy.util.exceptions import SunpyUserWarning


@pytest.fixture
def hgs_header(aia171_test_map):
    return sunpy.map.make_fitswcs_header(
        (180, 360),
        SkyCoord(0*u.deg, 0*u.deg,
                 frame='heliographic_stonyhurst',
                 obstime=aia171_test_map.date,
                 rsun=aia171_test_map.coordinate_frame.rsun),
        scale=(1, 1)*u.deg/u.pix,
        projection_code='CAR'
    )


@pytest.fixture
def hpc_header(aia171_test_map):
    new_observer = SkyCoord(45*u.deg, 0*u.deg, 1*u.AU,
                            frame='heliographic_stonyhurst',
                            obstime=aia171_test_map.date)
    return sunpy.map.make_fitswcs_header(
        aia171_test_map.data.shape,
        SkyCoord(0*u.arcsec, 0*u.arcsec,
                 frame='helioprojective',
                 obstime=aia171_test_map.date,
                 observer=new_observer,
                 rsun=aia171_test_map.coordinate_frame.rsun),
        scale=u.Quantity(aia171_test_map.scale),
        projection_code='TAN'
    )


@figure_test
def test_reproject_to_hgs(aia171_test_map, hgs_header):
    aia171_test_map.reproject_to(hgs_header).plot()


@check_figures_equal(extensions=["png"])
def test_reproject_to_hgs_wcs(fig_test, fig_ref, aia171_test_map, hgs_header):
    with warnings.catch_warnings():
        # NumPy <1.19 emits a RuntimeWarning because of comparison against NaNs
        warnings.filterwarnings("ignore", message='invalid value encountered',
                                category=RuntimeWarning)

        # Tests whether reprojecting to a WCS instance gives the same answer as to a header
        header_map = aia171_test_map.reproject_to(hgs_header)
        wcs_map = aia171_test_map.reproject_to(WCS(hgs_header))

        ax_ref = fig_ref.add_subplot(projection=header_map)
        header_map.plot(axes=ax_ref)

        ax_test = fig_test.add_subplot(projection=wcs_map)
        wcs_map.plot(axes=ax_test)


@check_figures_equal(extensions=["png"])
def test_reproject_to_hpc_default(fig_test, fig_ref, aia171_test_map, hpc_header):
    with warnings.catch_warnings():
        # NumPy <1.19 emits a RuntimeWarning because of comparison against NaNs
        warnings.filterwarnings("ignore", message='invalid value encountered',
                                category=RuntimeWarning)

        # Tests whether the default reprojection is "interpolation"
        default_map = aia171_test_map.reproject_to(hpc_header)
        interpolation_map = aia171_test_map.reproject_to(hpc_header, algorithm='interpolation')

        ax_ref = fig_ref.add_subplot(projection=interpolation_map)
        interpolation_map.plot(axes=ax_ref)

        ax_test = fig_test.add_subplot(projection=default_map)
        default_map.plot(axes=ax_test)


@figure_test
def test_reproject_to_hpc_interpolation(aia171_test_map, hpc_header):
    aia171_test_map.reproject_to(hpc_header, algorithm='interpolation').plot()


@figure_test
def test_reproject_to_hpc_exact(aia171_test_map, hpc_header):
    aia171_test_map.reproject_to(hpc_header, algorithm='exact').plot()


@figure_test
def test_reproject_to_hpc_adaptive(aia171_test_map, hpc_header):
    aia171_test_map.reproject_to(hpc_header, algorithm='adaptive', kernel='Hann', boundary_mode='strict').plot()


def test_return_footprint(aia171_test_map, hpc_header):
    with warnings.catch_warnings():
        # NumPy <1.19 emits a RuntimeWarning because of comparison against NaNs
        warnings.filterwarnings("ignore", message='invalid value encountered',
                                category=RuntimeWarning)

        return_without_footprint = aia171_test_map.reproject_to(hpc_header)
        assert isinstance(return_without_footprint, sunpy.map.GenericMap)

        return_with_footprint = aia171_test_map.reproject_to(hpc_header, return_footprint=True)
        assert len(return_with_footprint) == 2
        assert isinstance(return_with_footprint[0], sunpy.map.GenericMap)
        assert isinstance(return_with_footprint[1], np.ndarray)


def test_invalid_inputs(aia171_test_map, hpc_header):
    with pytest.raises(ValueError, match="The specified algorithm must be one of"):
        aia171_test_map.reproject_to(hpc_header, algorithm='something')


def test_rsun_mismatch_warning(aia171_test_map, hpc_header):
    # Modifying the `hpc_header` rsun value to create a mismatch
    hpc_header["rsun_ref"] += 1
    with pytest.warns(SunpyUserWarning, match="rsun mismatch detected: "):
        # Reproject with the mismatched rsun
        aia171_test_map.reproject_to(hpc_header)


@figure_test
@pytest.mark.parametrize('auto_extent', [None, 'corners', 'edges', 'all'])
def test_reproject_to_auto_extent(aia171_test_map, auto_extent):
    aia = aia171_test_map.submap([32, 32] * u.pix, top_right=[80, 100]*u.pix)

    header = {
        'naxis1': 60,
        'naxis2': 200,
        'crpix1': 30.5,
        'crpix2': 100.5,
        'crval1': -750,
        'crval2': 0,
        'cdelt1': 10,
        'cdelt2': 10,
        'cunit1': 'arcsec',
        'cunit2': 'arcsec',
        'ctype1': 'HPLN-TAN',
        'ctype2': 'HPLT-TAN',
        'hgln_obs': 80,
        'hglt_obs': 10,
        'rsun_ref': aia.rsun_meters.value,
        'dsun_obs': aia.dsun.value,
        'date-obs': aia.date.isot,
        'mjd-obs': aia.date.mjd,
    }

    reprojected_aia = aia.reproject_to(header, auto_extent=auto_extent)
    fig = Figure()
    ax = fig.add_subplot(projection=reprojected_aia)
    reprojected_aia.plot(axes=ax, title=f"auto_extent={auto_extent}")
    aia.draw_extent(axes=ax, color='red')
    return fig


@pytest.mark.parametrize(('auto_extent', 'crpix', 'pixel_shape'), [(None, [5.5, 5.5], [10, 10]),
                                                                   ('corners', [0.5, 836.5], [241, 1672]),
                                                                   ('edges', [0.5, 836.5], [482, 1672]),
                                                                   ('all', [0.5, 836.5], [963, 1672])])
def test_reproject_to_auto_extent_wcs(aia171_test_map, auto_extent, crpix, pixel_shape):
    hgs_header = {
        'naxis1': 15,
        'naxis2': 12,
        'crpix1': 0.5,
        'crpix2': 6.5,
        'crval1': 0,
        'crval2': 0,
        'cdelt1': 10,
        'cdelt2': 10,
        'cunit1': 'deg',
        'cunit2': 'deg',
        'ctype1': 'HGLN-CAR',
        'ctype2': 'HGLT-CAR',
        'rsun_ref': 700000000,
        'dsun_obs': 150000000000,
        'date-obs': '2001-01-01',
    }
    hgs_map = sunpy.map.Map((np.arange(180).reshape((12, 15)), hgs_header))

    hpc_header = {
        'naxis1': 10,
        'naxis2': 10,
        'crpix1': 5.5,
        'crpix2': 5.5,
        'crval1': 0,
        'crval2': 0,
        'cdelt1': 1,
        'cdelt2': 1,
        'cunit1': 'arcsec',
        'cunit2': 'arcsec',
        'ctype1': 'HPLN-TAN',
        'ctype2': 'HPLT-TAN',
        'hgln_obs': 0,
        'hglt_obs': 0,
        'rsun_ref': 700000000,
        'dsun_obs': 150000000000,
        'date-obs': '2001-01-01',
        'mjd-obs': 51910,
    }

    reprojected_wcs = hgs_map.reproject_to(hpc_header, auto_extent=auto_extent).wcs

    # These parts of the WCS should not change
    assert_quantity_allclose(reprojected_wcs.wcs.crval, hgs_map.wcs.wcs.crval)
    assert_quantity_allclose(reprojected_wcs.wcs.cdelt * u.deg, 1 * u.arcsec)

    # These parts of the WCS should have changed
    assert_quantity_allclose(reprojected_wcs.wcs.crpix, crpix)
    assert_quantity_allclose(reprojected_wcs.pixel_shape, pixel_shape)


@figure_test
@pytest.mark.parametrize("screen", [sunpy.coordinates.SphericalScreen, sunpy.coordinates.PlanarScreen])
def test_reproject_to_screen_plus_diffrot(aia171_test_map, screen):
    new_time = aia171_test_map.date + 5*u.day
    header = {
        'naxis1': 150,
        'naxis2': 150,
        'crpix1': 75.5,
        'crpix2': 75.5,
        'crval1': 0,
        'crval2': 0,
        'cdelt1': 20,
        'cdelt2': 20,
        'cunit1': 'arcsec',
        'cunit2': 'arcsec',
        'ctype1': 'HPLN-TAN',
        'ctype2': 'HPLT-TAN',
        'hgln_obs': 0,
        'hglt_obs': 30,
        'rsun_ref': aia171_test_map.rsun_meters.value,
        'dsun_obs': aia171_test_map.dsun.value,
        'date-obs': new_time.isot,
        'mjd-obs': new_time.mjd,
    }

    with sunpy.coordinates.transform_with_sun_center(), screen(aia171_test_map.observer_coordinate, only_off_disk=True):
        without_diffrot = aia171_test_map.reproject_to(header)
        with sunpy.coordinates.propagate_with_solar_surface():
            with_diffrot = aia171_test_map.reproject_to(header)

    fig = Figure(figsize=(11, 4))
    ax1 = fig.add_subplot(121, projection=without_diffrot)
    without_diffrot.plot(axes=ax1, title="Without diffrot")
    ax2 = fig.add_subplot(122, projection=with_diffrot)
    with_diffrot.plot(axes=ax2, title="With diffrot")
    return fig


@pytest.fixture
def m_eit_1():
    m = get_dummy_map_from_header('EIT_header/efz20040301.000010_s.header')
    # These EIT headers have no DATE-AVG key by default so we mimic it.
    m.meta['DATE-AVG'] = (m.date + m.exposure_time/2).isot
    return m


@pytest.fixture
def m_eit_1_no_date_avg():
    return get_dummy_map_from_header('EIT_header/efz20040301.000010_s.header')


@pytest.fixture
def m_eit_2():
    m = get_dummy_map_from_header('EIT_header/efz20040301.120010_s.header')
    m.meta['DATE-AVG'] = (m.date + m.exposure_time/2).isot
    return m


def test_reproject_to_preserve_date_obs(m_eit_1, m_eit_2):
    """Test that preserve_date_obs preserves the date of the original date in DATE-OBS"""
    with pytest.warns(SunpyUserWarning, match="rsun mismatch detected: "):
        m_2_r = m_eit_2.reproject_to(m_eit_1.wcs, preserve_date_obs=True)
    assert m_2_r.date == m_eit_2.date
    assert m_2_r.reference_date == m_eit_1.date_average
    assert m_2_r.coordinate_frame.obstime == m_eit_1.coordinate_frame.obstime


def test_reproject_to_preserve_date_obs_target_has_no_date_avg(m_eit_1_no_date_avg, m_eit_2):
    """Test that date is preserved even when target WCS does not have a date-avg"""
    with pytest.warns(SunpyUserWarning, match="rsun mismatch detected: "):
        m_2_r = m_eit_2.reproject_to(m_eit_1_no_date_avg.wcs, preserve_date_obs=True)
    assert m_2_r.date == m_eit_2.date
    assert m_2_r.reference_date == m_eit_1_no_date_avg.date
    assert m_2_r.coordinate_frame.obstime == m_eit_1_no_date_avg.coordinate_frame.obstime


def test_reproject_to_preserve_date_obs_raises_exception_missing_date_obs_avg(m_eit_1, m_eit_2):
    target_header = m_eit_1.wcs.to_header()
    for k in ['DATE-OBS', 'DATE-AVG', 'MJD-OBS', 'MJD-AVG']:
        target_header.remove(k)
    target_header.set('NAXIS1', value=m_eit_1.data.shape[1])
    target_header.set('NAXIS2', value=m_eit_1.data.shape[0])
    with pytest.raises(ValueError, match="The target_wcs has neither a DATE-AVG or DATE-OBS."):
        m_eit_2.reproject_to(target_header, preserve_date_obs=True)



import re
import string
import importlib
from abc import ABC, abstractmethod
from textwrap import dedent
from functools import wraps
from collections.abc import Sequence

import numpy as np

import astropy.units as u
from astropy.table import Column, QTable, Row, Table, TableAttribute

from sunpy.util.util import get_width

__all__ = ['QueryResponseColumn', 'BaseQueryResponse',
           'QueryResponseRow', 'QueryResponseTable', 'BaseClient',
           'convert_row_to_table']


class BaseQueryResponse(Sequence):
    """
    An Abstract Base Class for results returned from BaseClient.

    Notes
    -----
    * A QueryResponse or QueryResponseTable object must be able to be instantiated with only one
      iterable argument. (i.e. the ``__init__`` must only have one required
      argument).
    * The `client` property must be settable.
    * The base class does not prescribe how you store the results from your
      client, only that it must be possible to represent them as an astropy
      table in the ``build_table`` method.
    * ``__getitem__`` **must** return an instance of the type it was called on.
      I.e. it must always return an object of ``type(self)``.

    """

    @abstractmethod
    def build_table(self):
        """
        Return an `astropy.table.Table` representation of the query response.
        """

    @property
    @abstractmethod
    def client(self):
        """
        An instance of `BaseClient` used to generate the results.

        Generally this is used to fetch the results later.

        .. note::

            In general, this doesn't have to be the same instance of
            ``BaseClient``, this is left to the client developer. If there is a
            significant connection overhead in creating an instance of a client
            you might want it to be the same instance as used for the search.
        """

    @client.setter
    @abstractmethod
    def client(self, value):
        pass

    @property
    @abstractmethod
    def blocks(self):
        """
        A `collections.abc.Sequence` object which contains the records
        contained within the Query Response.
        """

    def response_block_properties(self):
        """
        Returns a set of class attributes on all the response blocks.

        Returns
        -------
        s : `set`
            List of strings, containing attribute names in the response blocks.
        """
        return set()

    def __str__(self):
        """Print out human-readable summary of records retrieved"""
        return '\n'.join(self.build_table().pformat(show_dtype=False))

    def __repr__(self):
        """Print out human-readable summary of records retrieved"""
        return object.__repr__(self) + "\n" + str(self)

    def _repr_html_(self):
        return self.build_table()._repr_html_()

    def show(self, *cols):
        """
        Returns response tables with desired columns for the Query.

        Parameters
        ----------
        \\*cols : `tuple`
            Name of columns to be shown.

        Returns
        -------
        `astropy.table.Table`
            A table showing values for specified columns.
        """
        table = self.build_table()
        if len(cols) == 0:
            return table
        tablecols = table.columns
        valid_cols = [col for col in cols if col in tablecols]
        return table[valid_cols]


class QueryResponseRow(Row):
    """
    A row subclass which knows about the client of the parent table.
    """

    def as_table(self):
        """
        Return this Row as a length one Table
        """
        return self.table[self.index:self.index + 1]

    def get(self, key, default=None):
        """
        Extract a value from the row if the key is present otherwise return the value of ``default``
        """
        if key in self.colnames:
            return self[key]
        return default

    @property
    def response_block_map(self):
        """
        A dictionary designed to be used to format a filename.

        This takes all the columns in this Row and lower cases them and
        replaces spaces with underscores. Also removes any characters not
        allowed in Python identifiers.
        """
        def key_clean(key):
            key = re.sub(f'[{re.escape(string.punctuation)}]', '_', key)
            key = key.replace(' ', '_')
            key = ''.join(char for char in key
                          if char.isidentifier() or char.isnumeric())
            return key.lower()

        return {key_clean(key): value for key, value in zip(self.colnames, self)}


class QueryResponseColumn(Column):
    """
    A column subclass which knows about the client of the parent table.
    """

    def as_table(self):
        """
        Return this Row as a length one Table
        """
        return self.parent_table[(self.name,)]


class QueryResponseTable(QTable):
    __doc__ = QTable.__doc__

    Row = QueryResponseRow
    Column = QueryResponseColumn

    client = TableAttribute()
    display_keys = TableAttribute(default=slice(None))
    hide_keys = TableAttribute()
    errors = TableAttribute(default=[])

    size_column = None

    def unhide_columns(self):
        """
        Modify this table so that all columns are displayed.
        """
        self.display_keys = slice(None)
        self.hide_keys = None
        return self

    def _reorder_columns(self, first_columns, remove_empty=True):
        """
        Generate a new version of this table with ``first_columns`` at the start.

        Parameters
        ----------
        first_columns : list
            The column names to put at the start of the table.
        remove_empty : bool, optional
            Remove columns where all values are `None`.
            Defaults to ``True``.

        Returns
        -------
        new_table : QueryResponseTable
            A sliced version of this table instance so that the columns are
            reordered.
        """
        all_cols = list(self.colnames)
        first_names = [n for n in first_columns if n in all_cols]
        extra_cols = [col for col in all_cols if col not in first_names]
        all_cols = first_names + extra_cols
        new_table = self[[col for col in all_cols if self[col] is not None]]
        if remove_empty:
            empty_cols = [
                col.info.name for col in self.itercols()
                if col.info.dtype.kind == 'O' and all(val is None for val in col)
            ]
            new_table.remove_columns(empty_cols)

        return new_table

    @property
    def _display_table(self):
        """
        Apply the display_keys and hide_keys attributes to the table.

        This removes any keys in hide keys and then slices by any keys in
        display_keys to return the correct table.
        """

        keys = list(self.colnames)
        if self.hide_keys:
            # Index only the keys not in hide keys in order
            [keys.remove(key) for key in self.hide_keys if key in keys]

        if self.display_keys != slice(None):
            keys = [dk for dk in self.display_keys if dk in keys]

        table = self[keys]
        # The slicing operation resets display and hide keys to default, but we
        # have already applied it
        table.unhide_columns()

        return table

    def __str__(self):
        """Print out human-readable summary of records retrieved"""
        return '\n'.join(self._display_table.pformat(show_dtype=False))

    def __repr__(self):
        """Print out human-readable summary of records retrieved"""
        return object.__repr__(self) + "\n" + str(self._display_table)

    def _repr_html_(self):
        return QTable._repr_html_(self._display_table)

    def show(self, *cols):
        """
        Return a table with only ``cols`` present.

        If no ``cols`` are specified, all columns will be shown, including any
        hidden by default.

        This differs slightly from ``QueryResponseTable[cols]`` as it allows
        keys which are not in the table to be requested.
        """
        table = self.copy()
        table.unhide_columns()

        if len(cols) == 0:
            return table

        valid_cols = [col for col in cols if col in table.colnames]
        table = table[valid_cols]

        # The slicing operation resets display and hide keys to default, but we
        # want to bypass it here.
        table.unhide_columns()
        return table

    def path_format_keys(self):
        """
        Returns all the names that can be used to format filenames.

        Each one corresponds to a single column in the table, and the format
        syntax should match the dtype of that column, i.e. for a ``Time``
        object or a ``Quantity``.
        """
        rbp = set(self[0].response_block_map.keys())
        for row in self[1:]:
            rbp.intersection(row.response_block_map.keys())
        return rbp

    def total_size(self):
        """
        Returns the total size of all files in a query.

        Derived classes must set the 'size_column' class attribute to make use
        of this.
        """
        if self.size_column not in self.colnames:
            return np.nan * u.byte
        sizes = self[self.size_column]
        # Strip negative filesizes
        total = np.nansum(sizes[sizes > 0])
        if not (total > 0 * u.byte):
            return np.nan * u.byte
        # Find the first power of 3 below the total filesize
        power = 10**(np.floor(np.log10(total.to_value(u.byte)) // 3) * 3)
        # Create mapping from prefix value to prefix name
        prefix_dict = {p[2]: p[0][0] for p in u.core.si_prefixes}
        prefix_unit = u.Unit(f'{prefix_dict[power]}byte')
        return total.to(prefix_unit).round(3)


BaseQueryResponse.register(QueryResponseTable)


def convert_row_to_table(func):
    """
    A wrapper to convert any `~.QueryResponseRow` objects to `~.QueryResponseTable` objects.
    """
    @wraps(func)
    def wrapper(self, query_results, **kwargs):
        if isinstance(query_results, QueryResponseRow):
            query_results = query_results.as_table()
        return func(self, query_results, **kwargs)

    return wrapper


def _print_client(client, html=False, visible_entries=None):
    """
    Given a BaseClient instance will print out each registered attribute.

    Parameters
    ----------
    client : BaseClient
        The instance class to print for.
    html : bool
        Will return a html table instead.

    Returns
    -------
    `str`
        String with the client.
    """
    width = -1 if html else get_width()
    class_name = f"{client.__module__+'.' or ''}{client.__class__.__name__}"
    attrs = client.register_values()
    lines = []
    t = Table(names=["Attr Type", "Name", "Description"],
              dtype=["U80", "U80", "U80"])
    for client_key in attrs.keys():
        # Work around for * attrs having one length.
        if len(attrs[client_key]) == 1 and attrs[client_key][0] == "*":
            t.add_row((client_key.__name__, "All", "All valid values"))
            continue
        for name, desc in attrs[client_key]:
            t.add_row((client_key.__name__, name, desc))
    lines = [class_name, dedent(client.__doc__.partition("\n\n")[0])]
    if html:
        lines = [f"<p>{line}</p>" for line in lines]
    lines.extend(t.pformat(max_lines=visible_entries, show_dtype=False,
                               max_width=width, align="<", html=html))
    return '\n'.join(lines)


class BaseClient(ABC):
    """
    This defines the Abstract Base Class for each download client.

    The BaseClient has several abstract methods that ensure that any subclass enforces the bare minimum API.
    These are `search`, `fetch` and `_can_handle_query`.
    The last one ensures that each download client can be registered with Fido.

    Most download clients should subclass `~sunpy.net.dataretriever.GenericClient`.
    If the structure of `~sunpy.net.dataretriever.GenericClient`
    is not useful you should use `BaseClient`.
    `~sunpy.net.vso.VSOClient` and `~sunpy.net.jsoc.JSOCClient`
    are examples of download clients that subclass ``BaseClient``.
    """

    _registry = dict()

    def __init_subclass__(cls, *args, **kwargs):
        """
        An __init_subclass__ hook initializes all of the subclasses of a given class.
        So for each subclass, it will call this block of code on import.
        This replicates some metaclass magic without the need to be aware of metaclasses.
        Here we use this to register each subclass in a dict that has the `_can_handle_query` attribute.
        This is then passed into the UnifiedDownloaderFactory so we can register them.
        This means that Fido can use the clients internally.
        """
        super().__init_subclass__(**kwargs)

        # We do not want to register GenericClient since its a dummy client.
        if cls.__name__ in ('GenericClient'):
            return

        cls._registry[cls] = cls._can_handle_query

        if hasattr(cls, "_attrs_module"):
            from sunpy.net import attrs

            name, module = cls._attrs_module()
            module_obj = importlib.import_module(module)

            existing_mod = getattr(attrs, name, None)
            if existing_mod and existing_mod is not module_obj:
                raise NameError(f"{name} has already been registered as an attrs name.")

            setattr(attrs, name, module_obj)

            if name not in attrs.__all__:
                attrs.__all__.append(name)

        # Register client attrs after it has registered its own attrs
        from sunpy.net import attr
        values = cls.register_values()
        # If the client has no support, we won't try to register attrs
        if values:
            attr.Attr.update_values({cls: values})

    def __repr__(self):
        """
        Returns the normal repr plus the pretty client __str__.
        """
        return object.__repr__(self) + "\n" + _print_client(visible_entries=15, client=self)

    def __str__(self):
        """
        This enables the "pretty" printing of BaseClient.
        """
        return _print_client(client=self)

    def _repr_html_(self):
        """
        This enables the "pretty" printing of the BaseClient with html.
        """
        return _print_client(visible_entries=15, client=self, html=True)

    @abstractmethod
    def search(self, *args, **kwargs):
        """
        This enables the user to search for data using the client.

        Must return a subclass of `BaseQueryResponse`.
        """

    @abstractmethod
    def fetch(self, query_results, *, path, downloader, **kwargs):
        """
        This enables the user to fetch the data using the client, after a search.

        Parameters
        ----------
        query_results:
            Results to download.
        path : `str` or `pathlib.Path`, optional
            Path to the download directory
        downloader : `parfive.Downloader`
            The download manager to use.

        Returns
        -------
        `parfive.Results`
            The results object, can be `None` if ``wait`` is `False` and
            ``downloader`` is not None.
        """

    @classmethod
    @abstractmethod
    def _can_handle_query(cls, *query):
        """
        This enables the client to register what kind of searches it can handle, to prevent Fido
        using the incorrect client.
        """

    @property
    def info_url(self):
        """
        This should return a string that is a URL to the data server or
        documentation on the data being served.
        """

    @staticmethod
    def check_attr_types_in_query(query, required_attrs={}, optional_attrs={}):
        """
        Check a query against required and optional attributes.

        Returns `True` if *query* contains all the attrs in *required_attrs*,
        and if *query* contains only attrs in both *required_attrs* and *optional_attrs*.
        """
        query_attrs = {type(x) for x in query}
        all_attrs = required_attrs.union(optional_attrs)
        return required_attrs.issubset(query_attrs) and query_attrs.issubset(all_attrs)

    @classmethod
    def register_values(cls, *query):
        """
        This enables the client to register what kind of Attrs it can use directly.

        Returns
        -------
        `dict`
            A dictionary with key values of Attrs and the values are a tuple of
            ("Attr Type", "Name", "Description").
        """
        return {}



"""
========================
Reading ADAPT FITS Files
========================

This example demonstrates how to load data from the
Air Force Data Assimilative Photospheric Flux Transport (ADAPT) model into a list of `sunpy.map.Map` objects.
"""
import matplotlib.pyplot as plt

import astropy.units as u
from astropy.io import fits

import sunpy.map
from sunpy.coordinates.sun import carrington_rotation_time
from sunpy.net import Fido
from sunpy.net import attrs as a

###############################################################################
# First we will download an ADAPT FITS file.
# To do this we will use the `sunpy.net.Fido` search interface to search for
# ADAPT data for Carrington Rotation 2193, and the first longitude type (0)
# which means the data will be in a Carrington coordinate frame.

date_start = carrington_rotation_time(2193)
date_end = date_start + 6*u.h

result = Fido.search(a.Time(date_start, date_end), a.Instrument('adapt'), a.adapt.ADAPTLonType("0"))
print(result)

downloaded_file = Fido.fetch(result)

###############################################################################
# ADAPT FITS files contain 12 realizations of synoptic magnetogram
# output as a result of varying model assumptions. `This is explained in detail in this talk
# <https://www.swpc.noaa.gov/sites/default/files/images/u33/SWW_2012_Talk_04_27_2012_Arge.pdf>`__.
#
# Because the array in the FITS file is 3D, it cannot be passed directly to `sunpy.map.Map`,
# as this will take the first slice only, ignoring the other realizations.
# Instead, we'll open the FITS file with `astropy.io.fits` and manually pull out each
# model realization.

adapt_fits = fits.open(downloaded_file[0])

###############################################################################
# ``adapt_fits`` is a list of ``HDU`` objects. The first of these contains
# the 12 realizations of the data and a header with sufficient information to build
# a list of maps. We unpack this information into a list of
# ``(data, header)`` tuples where ``data`` are the different adapt realizations.

data_header_pairs = [(map_slice, adapt_fits[0].header) for map_slice in adapt_fits[0].data]

###############################################################################
# Next, pass this list of tuples to `sunpy.map.Map` to make a list of maps:

adapt_maps = sunpy.map.Map(data_header_pairs)

###############################################################################
# ``adapt_maps`` is now a list of our individual adapt realizations.
# Here, we generate a static plot accessing a subset of the individual maps in turn.

fig = plt.figure(figsize=(5, 10), layout='constrained')
for i, a_map in enumerate(adapt_maps[::4]):
    ax = fig.add_subplot(3, 1, i+1, projection=a_map)
    ax.tick_params(axis='x', labelsize=6)
    im = a_map.plot(axes=ax, cmap='bwr', vmin=-50, vmax=50, title=f"Realization {1+i*4:02d}")

fig.colorbar(im, label='Magnetic Field Strength [G]', orientation='horizontal')

plt.show()



"""
==============================================
Rotating HMI maps so they're not 'upside-down'
==============================================

This example shows how to rotate a HMI magnetogram, so when you plot it
it appears with solar North pointing up.
"""
import matplotlib.pyplot as plt

import sunpy.map
from sunpy.data.sample import HMI_LOS_IMAGE

###############################################################################
# We will use the ``sunpy`` HMI sample image, load it into a map and plot it.
# We see that solar North is pointed down instead of up in this image, which is
# indicated by the coordinates (that range from positive to negative, rather
# than negative to positive).

hmi_map = sunpy.map.Map(HMI_LOS_IMAGE)
fig = plt.figure()
ax = fig.add_subplot(projection=hmi_map)
hmi_map.plot(axes=ax)

plt.show()

###############################################################################
# Now rotate the image such that solar North is pointed up.
# We have to do this because the HMI instrument is mounted upside-down
# relative to the AIA instrument on the SDO satellite, which means most
# of the images are taken with solar North pointed down.
# The roll angle of the instrument is reported in the FITS header
# keyword ``CROTA2`` (see Figure 17 of :cite:t:`couvidat_hmi_2016`
# which states that "the nominal CROTA2 for HMI is ≈179.93").
#
# The order keyword, below, specifies the type of interpolation;
# in this case, 3 refers to bi-cubic.

hmi_rotated = hmi_map.rotate(order=3)
fig = plt.figure()
ax = fig.add_subplot(projection=hmi_rotated)
hmi_rotated.plot(axes=ax)

plt.show()



"""
This module FERMI GBM `~sunpy.timeseries.TimeSeries` source.
"""
from collections import OrderedDict

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import astropy.units as u
from astropy.time import TimeDelta

import sunpy.io
import sunpy.io._file_tools
from sunpy.time import parse_time
from sunpy.timeseries.timeseriesbase import GenericTimeSeries
from sunpy.util.metadata import MetaDict
from sunpy.visualization import peek_show

__all__ = ['GBMSummaryTimeSeries']


class GBMSummaryTimeSeries(GenericTimeSeries):
    """
    Fermi/GBM Summary lightcurve TimeSeries.

    The Gamma-ray Burst Monitor (GBM) is an instrument on board Fermi.
    It is meant to detect gamma-ray bursts but also detects solar flares.
    It consists of 12 Sodium Iodide (NaI) scintillation detectors and 2 Bismuth Germanate (BGO) scintillation detectors.
    The NaI detectors cover from a few keV to about 1 MeV and provide burst triggers and locations.
    The BGO detectors cover the energy range from about 150 keV to about 30 MeV.

    This summary lightcurve makes use of the CSPEC (daily version) data set which consists of the counts
    accumulated every 4.096 seconds in 128 energy channels for each of the 14 detectors.

    Note that the data is re-binned from the original 128 into the following 8 pre-determined energy channels.
    The rebinning method treats the counts in each of the original 128 channels as
    all having the energy of the average energy of that channel. For example, the
    counts in an 14.5--15.6 keV original channel would all be accumulated into the
    15--25 keV rebinned channel.

    * 4-15 keV
    * 15-25 keV
    * 25-50 keV
    * 50-100 keV
    * 100-300 keV
    * 300-800 keV
    * 800-2000 keV

    Examples
    --------
    >>> import sunpy.timeseries
    >>> import sunpy.data.sample  # doctest: +REMOTE_DATA
    >>> gbm = sunpy.timeseries.TimeSeries(sunpy.data.sample.GBM_TIMESERIES, source='GBMSummary')  # doctest: +REMOTE_DATA
    >>> gbm.peek()   # doctest: +SKIP

    References
    ----------
    * `Fermi Mission Homepage <https://fermi.gsfc.nasa.gov>`__
    * `Fermi GBM Homepage <https://fermi.gsfc.nasa.gov/science/instruments/gbm.html>`__
    * `Fermi Science Support Center <https://fermi.gsfc.nasa.gov/ssc/>`__
    * `Fermi Data Product <https://fermi.gsfc.nasa.gov/ssc/data/access/>`__
    * `GBM Instrument Papers <https://gammaray.nsstc.nasa.gov/gbm/publications/instrument_journal_gbm.html>`__
    """
    # Class attributes used to specify the source class of the TimeSeries
    # and a URL to the mission website.
    _source = 'gbmsummary'
    _url = "https://gammaray.nsstc.nasa.gov/gbm/#"

    def plot(self, axes=None, columns=None, **kwargs):
        """
        Plots the GBM timeseries.

        Parameters
        ----------
        axes : `matplotlib.axes.Axes`, optional
            The axes on which to plot the TimeSeries. Defaults to current axes.
        columns : list[str], optional
            If provided, only plot the specified columns.
        **kwargs : `dict`
            Additional plot keyword arguments that are handed to `~matplotlib.axes.Axes.plot`
            functions.

        Returns
        -------
        `~matplotlib.axes.Axes`
            The plot axes.
        """
        axes, columns = self._setup_axes_columns(axes, columns)
        for d in columns:
            axes.plot(self._data.index, self._data[d], label=d, **kwargs)
        axes.set_yscale("log")
        axes.set_ylabel('Counts/s/keV')
        axes.legend()
        self._setup_x_axis(axes)
        return axes

    @peek_show
    def peek(self, *, title=None, columns=None, **kwargs):
        """
        Displays the GBM timeseries by calling
        `~sunpy.timeseries.sources.fermi_gbm.GBMSummaryTimeSeries.plot`.

        .. plot::

            import sunpy.timeseries
            import sunpy.data.sample
            gbm = sunpy.timeseries.TimeSeries(sunpy.data.sample.GBM_TIMESERIES, source='GBMSummary')
            gbm.peek()

        Parameters
        ----------
        title : `str`, optional
            The title of the plot.
        columns : list[str], optional
            If provided, only plot the specified columns.
        **kwargs : `dict`
            Additional plot keyword arguments that are handed to `~matplotlib.axes.Axes.plot`
            functions.
        """
        if title is None:
            title = 'Fermi GBM Summary data ' + str(self.meta.get('DETNAM').values())
        fig, ax = plt.subplots()
        axes = self.plot(axes=ax, columns=columns, **kwargs)
        axes.set_title(title)
        return fig

    @classmethod
    def _parse_file(cls, filepath):
        """
        Parses a GBM CSPEC FITS file.

        Parameters
        ----------
        filepath : `str`
            The path to the file you want to parse.
        """
        hdus = sunpy.io._file_tools.read_file(filepath)
        return cls._parse_hdus(hdus)

    @classmethod
    def _parse_hdus(cls, hdulist):
        """
        Parses a GBM CSPEC `astropy.io.fits.HDUList`.

        Parameters
        ----------
        hdulist : `str`
            The path to the file you want to parse.
        """
        header = MetaDict(OrderedDict(hdulist[0].header))
        # these GBM files have three FITS extensions.
        # extn1 - this gives the energy range for each of the 128 energy bins
        # extn2 - this contains the data, e.g. counts, exposure time, time of observation
        # extn3 - eclipse times?
        energy_bins = hdulist[1].data
        count_data = hdulist[2].data

        # rebin the 128 energy channels into some summary ranges
        # 4-15 keV, 15 - 25 keV, 25-50 keV, 50-100 keV, 100-300 keV, 300-800 keV, 800 - 2000 keV
        # put the data in the units of counts/s/keV
        summary_counts = _bin_data_for_summary(energy_bins, count_data)

        # get the time information in datetime format with the correct MET adjustment
        met_ref_time = parse_time('2001-01-01 00:00')  # Mission elapsed time
        gbm_times = met_ref_time + TimeDelta(count_data['time'], format='sec')
        gbm_times.precision = 9
        gbm_times = gbm_times.isot.astype('datetime64')

        column_labels = ['4-15 keV', '15-25 keV', '25-50 keV', '50-100 keV',
                         '100-300 keV', '300-800 keV', '800-2000 keV']

        # Add the units data
        units = OrderedDict([('4-15 keV', u.ct / u.s / u.keV), ('15-25 keV', u.ct / u.s / u.keV),
                             ('25-50 keV', u.ct / u.s / u.keV), ('50-100 keV', u.ct / u.s / u.keV),
                             ('100-300 keV', u.ct / u.s / u.keV), ('300-800 keV', u.ct / u.s / u.keV),
                             ('800-2000 keV', u.ct / u.s / u.keV)])
        return pd.DataFrame(summary_counts, columns=column_labels, index=gbm_times), header, units

    @classmethod
    def is_datasource_for(cls, **kwargs):
        """
        Determines if the file corresponds to a GBM summary lightcurve
        `~sunpy.timeseries.TimeSeries`.
        """
        # Check if source is explicitly assigned
        if 'source' in kwargs.keys():
            if kwargs.get('source', ''):
                return kwargs.get('source', '').lower().startswith(cls._source)
        # Check if HDU defines the source instrument
        if 'meta' in kwargs.keys():
            return kwargs['meta'].get('INSTRUME', '').startswith('GBM')


def _bin_data_for_summary(energy_bins, count_data):
    """
    Rebin the 128 energy channels into some summary ranges and put the data in
    the units of counts/s/keV.

    Bin ranges used:
    * 4-15 keV
    * 15-25 keV
    * 25-50 keV
    * 50-100 keV
    * 100-300 keV
    * 300-800 keV
    * 800-2000 keV

    Parameters
    ----------
    energy_bins : `numpy.ndarray`
        The array of energy bins to rebin.
    count_data : `numpy.ndarray`
        The array of count data to rebin.
    """

    # list of energy bands to sum between
    ebands = [4, 15, 25, 50, 100, 300, 800, 2000]
    e_center = (energy_bins['e_min'] + energy_bins['e_max']) / 2
    indices = [np.searchsorted(e_center, e) for e in ebands]

    summary_counts = []
    for ind_start, ind_end in zip(indices[:-1], indices[1:]):
        # sum the counts in the energy bands, and find counts/s/keV
        summed_counts = np.sum(count_data["counts"][:, ind_start:ind_end], axis=1)
        energy_width = (energy_bins["e_max"][ind_end - 1] - energy_bins["e_min"][ind_start])
        summary_counts.append(summed_counts/energy_width/count_data["exposure"])

    return np.array(summary_counts).T



import pytest

import astropy.units as u

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.sources.gong import GONGSynopticMap
from .helpers import _test_private_date_setters


@pytest.fixture
def gong_synoptic():
    return get_dummy_map_from_header(get_test_filepath('gong_synoptic.header'))


def test_fitstoGONGSynoptic(gong_synoptic):
    """Tests the creation of GongSynopticMap using FITS."""
    assert isinstance(gong_synoptic, GONGSynopticMap)


def test_is_datasource_for(gong_synoptic):
    """Test the is_datasource_for method of GongSynopticMap."""
    assert gong_synoptic.is_datasource_for(gong_synoptic.data, gong_synoptic.meta)


def test_observatory(gong_synoptic):
    """Tests the observatory property of the GongSynopticMap object."""
    assert gong_synoptic.observatory == "NSO-GONG"


def test_measurement(gong_synoptic):
    """Tests the measurement property of the GongSynopticMap object."""
    assert gong_synoptic.measurement == 676.8 * u.nm


def test_reference_date(gong_synoptic):
    assert gong_synoptic.reference_date.isot == "2023-09-30T06:44:00.000"


def test_date(gong_synoptic):
    assert gong_synoptic.date.isot == "2023-09-30T06:44:00.000"


def test_private_date_setters(gong_synoptic):
    _test_private_date_setters(gong_synoptic)


def test_unit(gong_synoptic):
    assert gong_synoptic.unit == u.G
    assert gong_synoptic.unit == u.Unit("Mx/cm^2")
    assert gong_synoptic.unit.to_string() == 'G'


def test_spatial_units(gong_synoptic):
    assert gong_synoptic.spatial_units[0] == u.deg
    assert gong_synoptic.spatial_units[1] == u.deg


def test_wcs(gong_synoptic):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    gong_synoptic.pixel_to_world(0*u.pix, 0*u.pix)



"""
This module provides NOAA Solar Cycle `~sunpy.timeseries.TimeSeries` source.
"""
from pathlib import Path
from collections import OrderedDict

import numpy as np
import pandas as pd

import astropy.units as u

from sunpy.timeseries.timeseriesbase import GenericTimeSeries
from sunpy.util.metadata import MetaDict
from sunpy.visualization import peek_show

__all__ = ['NOAAIndicesTimeSeries', 'NOAAPredictIndicesTimeSeries']


class NOAAIndicesTimeSeries(GenericTimeSeries):
    """
    NOAA Solar Cycle monthly indices.

    Solar activity is measured by a number of different values.
    The NOAA Solar Weather Prediction Center (SWPC) publishes the following indices.
    All of these indices are also provided as a 13-month running smoothed value.

    * The SWO sunspot number is issued by the NOAA Space Weather Prediction Center (SWPC)
    * The RI sunspot number is the official International Sunspot Number and is
      issued by the `Solar Influence Data Analysis Center (SDIC)
      <https://sidc.be/aboutSIDC/>`__ in Brussels, Belgium.
    * The ratio between the SWO and RI indices.
    * Radio flux at 10.7 cm is produced by
      `Penticon/Ottawa <https://www.ngdc.noaa.gov/stp/solar/flux.html>`__ and the units are in sfu.
    * The Ap Geomagnetic Index is produced by the United States Air Force (USAF).

    .. note::
        See the gallery example :ref:`sphx_glr_generated_gallery_plotting_solar_cycle_example.py`
        for how to use `~sunpy.net.Fido` to retrieve the data file.

    Examples
    --------
    >>> import sunpy.timeseries
    >>> noaa_url = "https://services.swpc.noaa.gov/json/solar-cycle/observed-solar-cycle-indices.json"
    >>> noaa = sunpy.timeseries.TimeSeries(noaa_url, source='NOAAIndices')  # doctest: +REMOTE_DATA
    >>> noaa.peek()  # doctest: +SKIP

    References
    ----------
    * `Solar and Geomagnetic Indices Data Archive <https://www.swpc.noaa.gov/products/3-day-geomagnetic-forecast>`__
    * `Recent solar indices <https://services.swpc.noaa.gov/json/solar-cycle/observed-solar-cycle-indices.json>`__
    * `Indices Descriptions <ftp://ftp.swpc.noaa.gov/pub/weekly/README3>`__
    * `NOAA plots of Solar Cycle Progression <https://www.swpc.noaa.gov/products/solar-cycle-progression>`__
    * `NOAA Product List <https://www.swpc.noaa.gov/products-and-data>`__
    """
    # Class attributes used to specify the source class of the TimeSeries
    # and a URL to the mission website.
    _source = 'noaaindices'
    _url = "https://www.swpc.noaa.gov/products/solar-cycle-progression"

    def plot(self, axes=None, plot_type='sunspot SWO', columns=None, **kwargs):
        """
        Plots NOAA Indices.

        Parameters
        ----------
        axes : `matplotlib.axes.Axes`, optional
            The axes on which to plot the TimeSeries.
        plot_type : {``"sunspot SWO"``, ``"sunspot RI"``, ``"sunspot compare"``, ``"radio"``, ``"geo"``}, optional
            The type of plot required. Defaults to ``"sunspot SWO"``.
        columns : list[str], optional
            Unused, but there to maintain uniformity among plot methods.
        **kwargs : `dict`
            Additional plot keyword arguments that are handed to `~matplotlib.axes.Axes.plot`
            functions.

        Returns
        -------
        `~matplotlib.axes.Axes`
            The plot axes.
        """
        axes, columns = self._setup_axes_columns(axes, columns)

        if plot_type == 'sunspot SWO':
            to_plot = ['sunspot SWO', 'sunspot SWO smooth']
            ylabel = 'Sunspot Number'
        elif plot_type == 'sunspot RI':
            to_plot = ['sunspot RI', 'sunspot RI smooth']
            ylabel = 'Sunspot Number'
        elif plot_type == 'sunspot compare':
            to_plot = ['sunspot RI', 'sunspot SWO']
            ylabel = 'Sunspot Number'
        elif plot_type == 'radio':
            to_plot = ['radio flux', 'radio flux smooth']
            ylabel = 'Radio Flux [sfu]'
        elif plot_type == 'geo':
            to_plot = ['geomagnetic ap', 'geomagnetic ap smooth']
            ylabel = 'Geomagnetic AP Index'
        else:
            raise ValueError(f'Got unknown plot type "{type}"')

        to_plot = self.to_dataframe()[to_plot]
        to_plot.plot(ax=axes, **kwargs)

        axes.set_ylim(0)
        axes.set_ylabel(ylabel)
        axes.legend()
        self._setup_x_axis(axes)
        return axes

    @peek_show
    def peek(self, *, title="Solar Cycle Progression", plot_type='sunspot SWO', columns=None, **kwargs):
        """
        Displays the NOAA Indices as a function of time by calling
        `~sunpy.timeseries.sources.noaa.NOAAPredictIndicesTimeSeries.plot`.

        .. plot::

            import sunpy.timeseries
            noaa_url = "https://services.swpc.noaa.gov/json/solar-cycle/observed-solar-cycle-indices.json"
            noaa = sunpy.timeseries.TimeSeries(noaa_url, source='NOAAIndices')
            noaa.peek()

        Parameters
        ----------
        title : `str`, optional
            The title of the plot. Defaults to "Solar Cycle Progression".
        plot_type : {``"sunspot SWO"``, ``"sunspot RI"``, ``"sunspot compare"``, ``"radio"``, ``"geo"``}, optional
            The type of plot required. Defaults to ``"sunspot SWO"``.
        columns : list[str], optional
            Unused, but there to maintain uniformity among peek methods.
        **kwargs : `dict`
            Additional plot keyword arguments that are handed to `~matplotlib.axes.Axes.plot`
            functions.
        """
        axes = self.plot(plot_type=plot_type, **kwargs)
        axes.set_title(title)
        return axes.get_figure()

    @classmethod
    def _parse_file(cls, filepath):
        suffix = Path(filepath).suffix
        if suffix == '.json':
            return cls._parse_json_file(filepath)
        else:
            raise ValueError(f"{Path(filepath).name} does not have a suffix of '.json'")

    @staticmethod
    def _parse_json_file(filepath):
        """
        Parses an NOAA indices JSON file.

        Parameters
        ----------
        filepath : `str`
            The path to the file you want to parse.
        """
        data = pd.read_json(filepath)
        rename = {'ssn': 'sunspot RI',
                  'smoothed_ssn': 'sunspot RI smooth',
                  'observed_swpc_ssn': 'sunspot SWO',
                  'smoothed_swpc_ssn': 'sunspot SWO smooth',
                  'f10.7': 'radio flux',
                  'smoothed_f10.7': 'radio flux smooth'}
        data = data.rename(columns=rename)
        # -1 is used as a fill value, replace with NaN
        data = data.replace(-1, np.nan)
        # Convoluted time index handling
        data = data.set_index('time-tag')
        data.index = pd.DatetimeIndex(data.index.values)

        # Add the units data, reported in radio flux values (sfu) originally.
        units = OrderedDict([('sunspot RI', u.dimensionless_unscaled),
                             ('sunspot RI smooth', u.dimensionless_unscaled),
                             ('sunspot SWO', u.dimensionless_unscaled),
                             ('sunspot SWO smooth', u.dimensionless_unscaled),
                             ('radio flux', 1e-22*u.W/(u.m**2*u.Hertz)),
                             ('radio flux smooth', 1e-22*u.W/(u.m**2*u.Hertz))])
        return data, MetaDict({'comments': ""}), units

    @classmethod
    def is_datasource_for(cls, **kwargs):
        """
        Determines if header corresponds to an NOAA indices timeseries.
        """
        if kwargs.get('source', ''):
            return kwargs.get('source', '').lower().startswith(cls._source)


class NOAAPredictIndicesTimeSeries(GenericTimeSeries):
    """
    NOAA Solar Cycle Predicted Progression.

    The predictions are updated monthly and are produced by ISES. Observed
    values are initially the preliminary values which are replaced with the
    final values as they become available.

    The following predicted values are available.

    * The predicted RI sunspot number is the official International Sunspot
      Number and is issued by the `Solar Influence Data Analysis Center (SDIC) <https://sidc.be/aboutSIDC/>`__ in Brussels, Belgium.
    * The predicted radio flux at 10.7 cm is produced by
      `Penticon/Ottawa <https://www.ngdc.noaa.gov/stp/solar/flux.html>`__ and the units are in sfu.

    .. note::
        See the gallery example :ref:`sphx_glr_generated_gallery_plotting_solar_cycle_example.py`
        for how to use `~sunpy.net.Fido` to retrieve the data file.

    Examples
    --------
    >>> import sunpy.timeseries
    >>> noaa_url = 'https://services.swpc.noaa.gov/json/solar-cycle/predicted-solar-cycle.json'  # doctest: +REMOTE_DATA
    >>> noaa = sunpy.timeseries.TimeSeries(noaa_url, source='NOAAPredictIndices')  # doctest: +REMOTE_DATA +IGNORE_WARNINGS
    >>> noaa.peek()  # doctest: +SKIP

    References
    ----------
    * `Solar and Geomagnetic Indices Data Archive <https://www.swpc.noaa.gov/products/3-day-geomagnetic-forecast>`__
    * `Predicted solar indices <https://services.swpc.noaa.gov/json/solar-cycle/predicted-solar-cycle.json>`__
    * `NOAA plots of Solar Cycle Progression <https://www.swpc.noaa.gov/products/solar-cycle-progression>`__
    * `NOAA Product List <https://www.swpc.noaa.gov/products-and-data>`__
    """

    # Class attributes used to specify the source class of the TimeSeries
    # and a URL to the mission website.
    _source = 'noaapredictindices'
    _url = "https://www.swpc.noaa.gov/products/solar-cycle-progression"

    _peek_title = "Solar Cycle Sunspot Number Prediction"

    def plot(self, axes=None, columns=None, **plot_args):
        """
        Plots predicted NOAA Indices as a function of time from a pandas dataframe.

        Parameters
        ----------
        axes : `matplotlib.axes.Axes`, optional
            The axes on which to plot the TimeSeries.
        columns : list[str], optional
            Unused, but there to maintain uniformity among plot methods.
        **kwargs : `dict`
            Additional plot keyword arguments that are handed to `~matplotlib.axes.Axes.plot`
            functions.

        Returns
        -------
        `~matplotlib.axes.Axes`
            The plot axes.
        """
        axes, columns = self._setup_axes_columns(axes, columns)

        dataframe = self.to_dataframe()
        dataframe['sunspot'].plot(color='b', **plot_args)
        dataframe['sunspot high'].plot(linestyle='--', color='b', **plot_args)
        dataframe['sunspot low'].plot(linestyle='--', color='b', **plot_args)
        axes.set_ylim(0)
        axes.set_ylabel('Sunspot Number')
        axes.yaxis.grid(True, 'major')
        axes.xaxis.grid(True, 'major')
        axes.legend()
        self._setup_x_axis(axes)
        return axes

    @classmethod
    def _parse_file(cls, filepath):
        suffix = Path(filepath).suffix
        if suffix == '.json':
            return cls._parse_json_file(filepath)
        else:
            raise ValueError(f"{Path(filepath).name} does not have a suffix of '.json'")

    @classmethod
    def is_datasource_for(cls, **kwargs):
        """
        Determines if header corresponds to an NOAA predict indices
        `~sunpy.timeseries.TimeSeries`.
        """
        if kwargs.get('source', ''):
            return kwargs.get('source', '').lower().startswith(cls._source)

    @staticmethod
    def _parse_json_file(filepath):
        """
        Parses an NOAA Predict indices JSON file.

        Parameters
        ----------
        filepath : `str`
            The path to the file you want to parse.
        """
        data = pd.read_json(filepath)
        rename = {'predicted_ssn': 'sunspot',
                  'high_ssn': 'sunspot high',
                  'low_ssn': 'sunspot low',
                  'predicted_f10.7': 'radio flux',
                  'high_f10.7': 'radio flux high',
                  'low_f10.7': 'radio flux low'}
        data = data.rename(columns=rename)
        # -1 is used as a fill value, replace with NaN
        data = data.replace(-1, np.nan)
        # Convoluted time index handling
        data = data.set_index('time-tag')
        data.index = pd.DatetimeIndex(data.index.values)
        # Add the units data, reported in radio flux values (sfu) originally.
        units = OrderedDict([('sunspot', u.dimensionless_unscaled),
                             ('sunspot high', u.dimensionless_unscaled),
                             ('sunspot low', u.dimensionless_unscaled),
                             ('radio flux', 1e-22*u.W/(u.m**2*u.Hertz)),
                             ('radio flux high', 1e-22*u.W/(u.m**2*u.Hertz)),
                             ('radio flux low', 1e-22*u.W/(u.m**2*u.Hertz))])
        return data, MetaDict({'comments': ""}), units
